#!/usr/bin/env python

import wx
import sys
import math
import six
import numpy as np
from wx import glcanvas
import OpenGL
OpenGL.ERROR_CHECKING = False
OpenGL.ERROR_LOGGING = False
OpenGL.ERROR_ON_COPY = True
from OpenGL.GL import *

vtxShaderSource = """
attribute vec3 VertexPosition;
attribute vec4 VertexColor;
uniform mat3 TransformationMatrix;
uniform vec3 VertexResolution;
uniform vec3 VertexScale;
uniform vec3 VertexScaleData;
uniform vec3 VertexOffset;
uniform vec3 VertexOffsetData;
uniform vec3 VertexScaleGL;
varying vec2 vTextureCoord;
varying vec4 vColor;
uniform int VertexOriginal;
void main(void) {
    if(VertexOriginal == 1) {
        gl_Position = vec4(VertexPosition, 1.0);
    } else if (VertexOriginal == 2) {
        gl_Position = vec4((((VertexPosition-VertexOffsetData)*VertexScaleData)*VertexScale+VertexOffset)/VertexResolution*2.0-1.0, 1.0);
        gl_Position = gl_Position*vec4(VertexScaleGL,1);
    } else if (VertexOriginal == 3) {
        gl_Position = vec4(((TransformationMatrix*((VertexPosition-VertexOffsetData)*VertexScaleData)*VertexScale)+VertexOffset)/VertexResolution*2.0-1.0, 1.0);
        gl_Position = gl_Position*vec4(VertexScaleGL,1);
        gl_Position = gl_Position + abs(gl_Position)*vec4(0, 0, -0.03, 0);
    } else {
        gl_Position = vec4(((TransformationMatrix*((VertexPosition-VertexOffsetData)*VertexScaleData)*VertexScale)+VertexOffset)/VertexResolution*2.0-1.0, 1.0);
        gl_Position = gl_Position*vec4(VertexScaleGL,1);
    }
    vColor = VertexColor;
}
"""

fragShaderSource = """
varying vec4 vColor;
uniform int mesh;
uniform vec4 clr;
uniform sampler2D uSampler;
void main(void) {
    if(mesh == 0) {
        gl_FragColor = vColor;
    } else if(mesh == 1) {
        gl_FragColor = vec4(0, 0, 1, 0.5);
    } else if(mesh == 2) {
        gl_FragColor = clr;
    } else if(mesh == 3) {
        vec4 textureColor = texture2D(uSampler, vec2(vColor[0], vColor[1]));
        gl_FragColor = vec4(textureColor.rgb , textureColor.a);
    }
}
"""


class SimpleBarBuf(object):
    def __init__(self, sz, rng, horz, clr=[0, 0.6, 0, 1]):
        self.horz = horz
        self.range = dict(rng)
        self.bar_color = clr[:]
        self.buf_size = -1
        self.Resize(sz)

    def SetRange(self, rng):
        if rng == self.range:
            return
        self.range.update(rng)
        self.Resize(self.buf_size)

    def SetColor(self, clr):
        if clr == self.bar_color:
            return
        self.bar_color = clr[:]
        if self.buf_size > 0:
            self.color = np.repeat(np.array([clr]), self.buf_size * 4,
                                   axis=0).flatten()

    def Resize(self, sz):
        self.buf_size = sz
        zmax = self.range['zmax']
        self.vertex = np.zeros((sz * 4, 3))
        self.vertex[:, 2].fill(zmax)
        if self.horz:
            # for point d[i], draw lines between points:
            # (i, ymax), (i, ymax-d[i]*g), (i+1, ymax-d[i]*g), (i+1, ymax)
            # see update() for detail
            xmax, xmin = self.range['xmax'], self.range['xmin']
            x = np.linspace(xmin, xmax, sz, endpoint=False).reshape((sz, 1))
            delta = (xmax - xmin) / sz
            self.vertex[:, 0] = (np.repeat(x, 4, axis=1) +
                                 [0, 0, delta, delta]).flatten()
        else:
            ymax, ymin = self.range['ymax'], self.range['ymin']
            y = np.linspace(ymin, ymax, sz, endpoint=False).reshape((sz, 1))
            delta = (ymax - ymin) / sz
            self.vertex[:, 1] = (np.repeat(y, 4, axis=1) +
                                 [0, 0, delta, delta]).flatten()

        self.color = np.repeat(np.array([self.bar_color]),
                               self.buf_size * 4,
                               axis=0).flatten()
        self.line = np.arange(sz * 4)

    def GetGLObject(self):
        return self.vertex.flatten(), self.color, self.line

    def SetData(self, d):
        if len(d) != self.buf_size:
            raise ValueError()
        zmax = self.range['zmax']
        if self.horz:
            ymax, ymin = self.range['ymax'], self.range['ymin']
            self.vertex[:, 1] = ymax - (np.kron(d, [0, 1, 1, 0])).flatten()
        else:
            xmax, xmin = self.range['xmax'], self.range['xmin']
            self.vertex[:, 0] = xmax - (np.kron(d, [0, 1, 1, 0])).flatten()


class SurfaceBase(glcanvas.GLCanvas):
    ID_SHOW_2D = wx.NewId()
    ID_SHOW_STEP_SURFACE = wx.NewId()
    ID_SHOW_TRIANGLE = wx.NewId()
    ID_SHOW_MESH = wx.NewId()
    ID_SHOW_CONTOUR = wx.NewId()
    ID_SHOW_BOX = wx.NewId()
    ID_SHOW_AXIS = wx.NewId()
    ID_SHOW_HORZ_BAR = wx.NewId()
    ID_SHOW_VERT_BAR = wx.NewId()
    ID_ROTATE_0 = wx.NewId()
    ID_ROTATE_90 = wx.NewId()
    ID_ROTATE_180 = wx.NewId()
    ID_ROTATE_270 = wx.NewId()

    def __init__(self, parent, points=None):
        # set the depth size, otherwise the depth test may not work correctly.
        attribs = [
            glcanvas.WX_GL_RGBA, glcanvas.WX_GL_DOUBLEBUFFER,
            glcanvas.WX_GL_DEPTH_SIZE, 16
        ]
        glcanvas.GLCanvas.__init__(self, parent, -1, attribList=attribs)
        self.context = glcanvas.GLContext(self)

        self.drag_start = wx.Point(0, 0)
        self.drag_end = wx.Point(0, 0)
        self.points = []
        self.pointsClr = []
        self.dimension = {'x': 0, 'y': 0}
        self.is_mouse_down = False
        self.is_dragging = False
        self.rotation = {'x': 110, 'y': 0, 'z': 110}
        self.default_rotate = 0
        sz = self.GetClientSize()
        self.W = sz.x
        self.H = sz.y
        self.range = {
            'xmin': 0,
            'ymin': 0,
            'zmin': 0,
            'xmax': 0,
            'ymax': 0,
            'zmax': 1
        }
        self.zrange = []
        self.scale = {'base': 1, 'zoom': 1}
        self.rotate_delta = 0.05
        self.margin = {'top': 50, 'bottom': 10, 'left': 1, 'right': 1}
        self.offset = {'base': {'x': 0, 'y': 0}, 'user': {'x': 0, 'y': 0}}
        self.data_offset = {'x': 0, 'y': 0, 'z': 0}
        self.selected = {'x': 0, 'y': 0, 'clr': [1, 0, 1, 1]}
        self.color_scale = []
        self.blocks = []
        self.data_zscale = 1
        # 2d mode is the fast-drawing mode, only shows the 2D image
        self.mode_2d = False
        self.show = {
            'surface': True,
            'mesh': False,
            'contour': False,
            'box': False,
            'axis': False,
            'horz_bar': False,
            'vert_bar': False
        }
        self.show_step_surface = True
        self.contour_levels = 100
        self.rotate_matrix = np.eye(3, dtype=np.float32)
        self.color_map = np.array(
            [[0, 0, 0, 1], [0, 0, 1, 1], [0, 1, 1, 1], [0, 1, 0, 1],
             [1, 1, 0, 1], [1, 0.5, 0, 1], [1, 0, 0, 1]], np.float32)
        self.initialized = False
        self.raw_points = points
        self.globjects = {}
        self.hudtext = ''
        self._hudtext = ''
        self._hudBuffer = np.zeros((0, 0, 4))
        # buffers to draw bar
        rows, cols = self.raw_points['z'].shape
        self.horzbar_buf = SimpleBarBuf(cols, self.range, True)
        self.vertbar_buf = SimpleBarBuf(rows, self.range, False)

        self.background_color = [0, 0, 0, 1]

        # Ugly. do not call SetImage or similar methods here, since it may be
        # overridden or may call some overridden methods. Instead, it should be
        # put in Initialize() method
        #self.SetImage(self.raw_points)

        accel_tbl = self.GetAccelList()
        self.SetAcceleratorTable(wx.AcceleratorTable(accel_tbl))

        self.Bind(wx.EVT_ERASE_BACKGROUND, self.OnEraseBackground)
        self.Bind(wx.EVT_SIZE, self.OnSize)
        self.Bind(wx.EVT_PAINT, self.OnPaint)
        self.Bind(wx.EVT_LEFT_DOWN, self.OnMouseDown)
        self.Bind(wx.EVT_LEFT_UP, self.OnMouseUp)
        self.Bind(wx.EVT_MOTION, self.OnMouseMotion)
        self.Bind(wx.EVT_KEY_DOWN, self.OnKeyDown)
        self.Bind(wx.EVT_LEFT_DCLICK, self.OnDoubleClick)
        self.Bind(wx.EVT_MOUSEWHEEL, self.OnMouseWheel)
        self.Bind(wx.EVT_CONTEXT_MENU, self.OnContextMenu)
        self.Bind(wx.EVT_MENU, self.OnProcessMenuEvent)
        self.Bind(wx.EVT_UPDATE_UI, self.OnUpdateMenu)

    def GetAccelList(self):
        accel_tbl = [
            (wx.ACCEL_SHIFT, ord('S'), self.ID_SHOW_TRIANGLE),
            (wx.ACCEL_SHIFT, ord('H'), self.ID_SHOW_HORZ_BAR),
            (wx.ACCEL_SHIFT, ord('V'), self.ID_SHOW_VERT_BAR),
        ]
        return accel_tbl

    def OnEraseBackground(self, event):
        pass  # Do nothing, to avoid flashing on MSW.

    def OnContextMenu(self, event):
        menu = self.GetContextMenu()
        if menu:
            self.PopupMenu(menu)

    def GetContextMenu(self):
        menu = wx.Menu()
        elements = wx.Menu()
        elements.Append(self.ID_SHOW_2D, 'Show 2D Mode', '', wx.ITEM_CHECK)
        elements.AppendSeparator()
        elements.Append(self.ID_SHOW_STEP_SURFACE, 'Step Surface', '',
                        wx.ITEM_CHECK)
        elements.Append(self.ID_SHOW_TRIANGLE, 'Show Surface\tshift+s', '',
                        wx.ITEM_CHECK)
        elements.Append(self.ID_SHOW_MESH, 'Show Mesh', '', wx.ITEM_CHECK)
        elements.AppendSeparator()
        elements.Append(self.ID_SHOW_CONTOUR, 'Show Contour', '',
                        wx.ITEM_CHECK)
        elements.Append(self.ID_SHOW_BOX, 'Show Box', '', wx.ITEM_CHECK)
        elements.Append(self.ID_SHOW_AXIS, 'Show Axis', '', wx.ITEM_CHECK)
        elements.AppendSeparator()
        elements.Append(self.ID_SHOW_HORZ_BAR, 'Show Horz Bar\tshift+h', '',
                        wx.ITEM_CHECK)
        elements.Append(self.ID_SHOW_VERT_BAR, 'Show Vert Bar\tshift+v', '',
                        wx.ITEM_CHECK)
        menu.AppendSeparator()
        menu.AppendSubMenu(elements, 'Elements')
        rotate = wx.Menu()
        rotate.Append(self.ID_ROTATE_0, 'Rotate 0 degree', '', wx.ITEM_CHECK)
        rotate.Append(self.ID_ROTATE_90, 'Rotate 90 degree', '', wx.ITEM_CHECK)
        rotate.Append(self.ID_ROTATE_180, 'Rotate 180 degree', '',
                      wx.ITEM_CHECK)
        rotate.Append(self.ID_ROTATE_270, 'Rotate 270 degree', '',
                      wx.ITEM_CHECK)
        menu.AppendSeparator()
        menu.AppendSubMenu(rotate, 'Orientation')
        return menu

    def SetSurfaceBackground(self, clr):
        self.background_color = clr
        self.Refresh()

    def GetSurfaceBackground(self):
        return self.background_color

    def Set2dMode(self, is2d):
        self.mode_2d = is2d
        if not self.mode_2d:
            self.SetImage(self.raw_points)

    def Get2dMode(self):
        return self.mode_2d

    def SetShowStepSurface(self, step):
        self.show_step_surface = step
        self.Invalidate()

    def GetShowStepSurface(self):
        return self.show_step_surface

    def SetShowMode(self, **kwargs):
        refresh = False
        genGLobj = False
        for key in six.iterkeys(kwargs):
            if key in self.show and self.show[key] != kwargs[key]:
                self.show[key] = kwargs[key]
                refresh = True
                if key in ['mesh', 'contour']:
                    genGLobj = True
        if genGLobj:
            self.Invalidate()
        if refresh:
            self.Refresh()

    def OnProcessMenuEvent(self, event):
        eid = event.GetId()
        show = {}
        if eid == self.ID_SHOW_2D:
            self.Set2dMode(not self.mode_2d)
            self.Refresh()
        elif eid == self.ID_SHOW_STEP_SURFACE:
            self.SetShowStepSurface(not self.GetShowStepSurface())
        elif eid == self.ID_SHOW_TRIANGLE:
            show['surface'] = not self.show['surface']
        elif eid == self.ID_SHOW_MESH:
            show['mesh'] = not self.show['mesh']
        elif eid == self.ID_SHOW_CONTOUR:
            show['contour'] = not self.show['contour']
        elif eid == self.ID_SHOW_BOX:
            show['box'] = not self.show['box']
        elif eid == self.ID_SHOW_AXIS:
            show['axis'] = not self.show['axis']
        elif eid == self.ID_SHOW_HORZ_BAR:
            show['horz_bar'] = not self.show['horz_bar']
        elif eid == self.ID_SHOW_VERT_BAR:
            show['vert_bar'] = not self.show['vert_bar']
        else:
            if eid == self.ID_ROTATE_0:
                self.default_rotate = 0
            elif eid == self.ID_ROTATE_90:
                self.default_rotate = 90
            elif eid == self.ID_ROTATE_180:
                self.default_rotate = 180
            elif eid == self.ID_ROTATE_270:
                self.default_rotate = 270
            self.ResetRotate()
            self.Resize()
            self.Refresh()
            return
        if show:
            self.SetShowMode(**show)

    def OnUpdateMenu(self, event):
        eid = event.GetId()
        if eid == self.ID_SHOW_2D:
            event.Check(self.mode_2d)
        elif eid == self.ID_SHOW_STEP_SURFACE:
            event.Check(self.GetShowStepSurface())
        elif eid == self.ID_SHOW_TRIANGLE:
            event.Check(self.show['surface'])
        elif eid == self.ID_SHOW_MESH:
            event.Check(self.show['mesh'])
            event.Enable(not self.mode_2d)
        elif eid == self.ID_SHOW_CONTOUR:
            event.Check(self.show['contour'])
            event.Enable(not self.mode_2d)
        elif eid == self.ID_SHOW_BOX:
            event.Check(self.show['box'])
            event.Enable(not self.mode_2d)
        elif eid == self.ID_SHOW_AXIS:
            event.Check(self.show['axis'])
            event.Enable(not self.mode_2d)
        elif eid == self.ID_SHOW_HORZ_BAR:
            event.Check(self.show['horz_bar'])
        elif eid == self.ID_SHOW_VERT_BAR:
            event.Check(self.show['vert_bar'])
        elif eid == self.ID_ROTATE_0:
            event.Check(self.default_rotate == 0)
        elif eid == self.ID_ROTATE_90:
            event.Check(self.default_rotate == 90)
        elif eid == self.ID_ROTATE_180:
            event.Check(self.default_rotate == 180)
        elif eid == self.ID_ROTATE_270:
            event.Check(self.default_rotate == 270)

    def SetMargin(self, left=None, right=None, top=None, bottom=None):
        for m in zip([left, right, top, bottom],
                     ['left', 'right', 'top', 'bottom']):
            if m[0] is not None:
                self.margin[m[1]] = m[0]

    def GetMargin(self):
        return self.margin

    def Resize(self):
        sz = self.GetClientSize()
        self.W = sz.x
        self.H = sz.y
        glViewport(0, 0, self.W, self.H)
        xmax, xmin = self.range['xmax'], self.range['xmin']
        ymax, ymin = self.range['ymax'], self.range['ymin']
        if self.default_rotate in [90, 270]:
            xmax, xmin, ymax, ymin = ymax, ymin, xmax, xmin
        if xmax - xmin > 0 and ymax - ymin >= 0:
            t, b = self.margin['top'], self.margin['bottom']
            l, r = self.margin['left'], self.margin['right']
            self.scale['base'] = min((self.W - l - r) / (xmax - xmin),
                                     (self.H - t - b) / (ymax - ymin))
            self.offset['base'] = {
                'x': self.W / 2 + (l - r) / 2,
                'y': self.H / 2 + (t - b) / 2
            }

    def OnSize(self, event):
        # update the size
        self.Resize()
        self.Refresh()
        event.Skip()

    def OnPaint(self, event):
        #dc = wx.PaintDC(self)
        self.SetCurrent(self.context)
        if not self.initialized:
            self.initialized = True
            self.Initialize()
        self.Draw()
        self.SwapBuffers()

    def Initialize(self):
        self.InitGL()
        if self.raw_points is not None:
            self.SetImage(self.raw_points)
            self.ResetRotate()

    def InitGL(self):
        self.SetCurrent(self.context)

        glClearColor(1.0, 1.0, 1.0, 1.0)  # this is the color
        glEnable(GL_DEPTH_TEST)  # Enable Depth Testing
        glDepthFunc(GL_LEQUAL)  # Set Perspective View
        glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA)
        glEnable(GL_BLEND)

        glEnable(GL_LINE_SMOOTH)

        FShader = fragShaderSource
        VShader = vtxShaderSource
        #Load and Compile Fragment Shader
        Code = FShader
        FShader = glCreateShader(GL_FRAGMENT_SHADER)
        glShaderSource(FShader, Code)
        glCompileShader(FShader)

        #Load and Compile Vertex Shader
        Code = VShader
        VShader = glCreateShader(GL_VERTEX_SHADER)
        glShaderSource(VShader, Code)
        glCompileShader(VShader)
        #print('cmp', glGetShaderiv(VShader, GL_COMPILE_STATUS))
        #print(glGetProgramInfoLog(FShader))
        #Create The Shader Program
        self.ShaderProgram = glCreateProgram()
        glAttachShader(self.ShaderProgram, FShader)
        glAttachShader(self.ShaderProgram, VShader)
        glLinkProgram(self.ShaderProgram)
        glUseProgram(self.ShaderProgram)
        #Link Vertex Position Attribute from Shader
        self.VertexPosition = glGetAttribLocation(self.ShaderProgram,
                                                  "VertexPosition")
        glEnableVertexAttribArray(self.VertexPosition)
        self.VertexColor = glGetAttribLocation(self.ShaderProgram,
                                               "VertexColor")
        glEnableVertexAttribArray(self.VertexColor)
        glViewport(0, 0, self.W, self.H)
        # generate the buffer
        self.glBufs = glGenBuffers(3)
        self.glTextures = glGenTextures(3)  # axis, hud, 2d image

    def OnKeyDown(self, event):
        if event.ShiftDown():
            if self.mode_2d:
                event.Skip()
                return
            keycode = event.GetKeyCode()
            x, y, z = 0, 0, 0
            if keycode == wx.WXK_UP:
                x = 1
            elif keycode == wx.WXK_DOWN:
                x = -1
            elif keycode == wx.WXK_LEFT:
                if event.CmdDown():
                    z = -1
                else:
                    y = -1
            elif keycode == wx.WXK_RIGHT:
                if event.CmdDown():
                    z = 1
                else:
                    y = 1
            else:
                event.Skip()
            self.Rotate(x, y, z)
        else:
            if self.dimension['x'] <= 0 or self.dimension['y'] <= 0:
                return
            keycode = event.GetKeyCode()
            x = self.selected['x']
            y = self.selected['y']
            mx = self.dimension['x'] - 1
            my = self.dimension['y'] - 1
            delta = 1
            dx, dy = 0, 0
            if event.CmdDown():
                delta = 5
            if keycode == wx.WXK_UP:
                dy = -delta
            elif keycode == wx.WXK_DOWN:
                dy = delta
            elif keycode == wx.WXK_LEFT:
                dx = -delta
            elif keycode == wx.WXK_RIGHT:
                dx = delta
            else:
                event.Skip()
                return
            # adjust the movement based on the current orientation
            if self.default_rotate == 90:
                dx, dy = dy, -dx
            elif self.default_rotate == 180:
                dx, dy = -dx, -dy
            elif self.default_rotate == 270:
                dx, dy = -dy, dx
            y = (y + dy) % my
            x = (x + dx) % mx
            self.SetSelected({'x': x, 'y': y})

    def OnMouseDown(self, event):
        bRect = self.GetClientRect()
        pos = event.GetPosition()
        self.drag_start.x = (pos.x - bRect.left)  #*(self.W/bRect.width)
        self.drag_start.y = (pos.y - bRect.top)  #*(self.H/bRect.height)
        self.is_mouse_down = True

    def OnMouseMotion(self, event):
        bRect = self.GetClientRect()
        pos = event.GetPosition()
        self.drag_end.x = (pos.x - bRect.left)  #*((self.W)/bRect.width)
        self.drag_end.y = (pos.y - bRect.top)  #*((self.H)/bRect.height)
        if self.is_mouse_down:
            if event.CmdDown():
                # when ctrl is down, move the image with mouse
                self.offset['user']['x'] += self.drag_end.x - self.drag_start.x
                self.offset['user']['y'] += self.drag_end.y - self.drag_start.y
                self.drag_start.x = self.drag_end.x
                self.drag_start.y = self.drag_end.y
                self.Refresh()
            elif not self.mode_2d:
                # rotate in 3d mode
                if abs(self.drag_start.x-self.drag_end.x) > 1 or \
                   abs(self.drag_start.y-self.drag_end.y) > 1:
                    self.is_dragging = True
                if self.is_dragging:
                    self.Rotate(-(self.drag_end.y - self.drag_start.y),
                                (self.drag_end.x - self.drag_start.x), 0)
                self.drag_start.x = self.drag_end.x
                self.drag_start.y = self.drag_end.y

    def OnMouseUp(self, event):
        self.is_mouse_down = False
        self.is_dragging = False
        self.Refresh()

    def OnDoubleClick(self, event):
        self.ResetRotate()
        self.Refresh()

    def OnMouseWheel(self, event):
        delta = (event.GetWheelRotation() > 0) * 2 - 1
        bRect = self.GetClientRect()
        pos = event.GetPosition()
        x = (pos.x - bRect.left) * ((self.W) / bRect.width)
        y = (pos.y - bRect.top) * ((self.H) / bRect.height)
        if delta:
            zoom = 1.1**delta
            self.scale['zoom'] *= zoom
            base = self.offset['base']
            user = self.offset['user']
            user['x'] = x - (x - base['x'] - user['x']) * zoom - base['x']
            user['y'] = y - (y - base['y'] - user['y']) * zoom - base['y']
            self.Refresh()

    def ResetRotate(self):
        # reset rotation, zoom, offset
        self.rotation = {'x': 0, 'y': 0, 'z': 0}
        self.rotate_matrix = np.eye(3, dtype=np.float32)
        self.scale['zoom'] = 1
        self.offset['user'] = {'x': 0, 'y': 0}
        # set the default orientation
        self.Rotate(0, 0,
                    self.default_rotate / 180. * np.pi / self.rotate_delta)

    def GetColorByZ(self, z):
        zMin = self.range['zmin']
        zMax = self.range['zmax']
        scale = self.data_zscale
        if zMax - zMin > 0:
            scale = 1 / (zMax - zMin)
        offset = 0
        if len(self.color_scale) == 2:
            scale = self.color_scale[0]
            offset = self.color_scale[1]

        C = self.color_map
        norm_p = (z - zMin) * scale + offset
        norm_p = norm_p * len(self.color_map)
        norm_p = np.clip(norm_p, 0, len(self.color_map) - 1)
        norm_p_l = np.floor(norm_p).astype(int)
        norm_p_r = np.ceil(norm_p).astype(int)
        norm_p_d = norm_p - norm_p_l
        return C[norm_p_l, :] + ((
            (C[norm_p_r, :] - C[norm_p_l, :]).T) * norm_p_d).T

    def Colorize(self):
        self.pointsClr = self.GetColorByZ(self.points[:, 2])
        sel = self.selected['x'] * self.dimension['y'] + self.selected['y']
        if sel >= 0 and sel < self.points.shape[0]:
            self.pointsClr[sel, :] = self.selected['clr']

    def SetContourLevels(self, level):
        if level == self.contour_levels:
            return
        self.contour_levels = level

        # the objects will be re-generated before drawing
        self.globject = None
        self.Refresh()

    def GetContourLevels(self):
        return self.contour_levels

    def SetDataScaleZ(self, scale):
        # add additional scale to z value, so that it may look better when
        # rotate the image (to project z value on x/y axis); otherwise, the
        # project may be too small or large. For example, x/y: [0: 100],
        # z [0: 0.01]
        self.data_zscale = scale

    def SetRange(self, rng):
        self.range.update(rng)
        #self.SetRangeZ(self.zrange)
        self.horzbar_buf.SetRange(self.range)
        self.vertbar_buf.SetRange(self.range)
        self.Resize()
        self._UpdateDataSacleZ()
        self.Invalidate()

    def GetRange(self):
        return self.range

    def _UpdateDataSacleZ(self):
        if len(self.zrange) == 2:
            zmin, zmax = self.zrange
        elif self.raw_points:
            z = self.raw_points['z']
            zmin, zmax = np.min(z), np.max(z)
        else:
            zmin, zmax = 0, 1
        self.data_offset['z'] = (zmax + zmin) / 2
        xmax, xmin = self.range['xmax'], self.range['xmin']
        ymax, ymin = self.range['ymax'], self.range['ymin']
        rng = max(ymax - ymin, xmax - xmin)
        # scale the z data, such that it has similar span as x or y axis.
        # otherwise, when rotate the image, the projection of z on x or y axis
        # will not look good (e.g., too small or too large).
        zscale = 1
        if zmax - zmin > 0 and rng > 0:
            zscale = float(rng) / (zmax - zmin)
        self.SetDataScaleZ(zscale)

    def SetRangeZ(self, zrange):
        # Set the range of the image values; otherwise it is calculated from
        # the data itself. It will impact the color of each pixel. Instead of
        # fully utilizing all the colors in color map, we may need to fix the
        # color code for each value (e.g., 0 is always blue)
        self.zrange = []
        if len(zrange) == 2:
            self.zrange = zrange
            self.SetRange({'zmin': zrange[0], 'zmax': zrange[1]})

    def SetHudText(self, txt):
        if txt != self.hudtext:
            self.hudtext = txt
            self.Refresh()

    def GetHudText(self):
        return self.hudtext

    def UpdateHudText(self):
        x, y = self.selected['x'], self.selected['y']
        r, c = self.raw_points['z'].shape
        if x >= 0 and y >= 0 and x < c and y < r:
            self.SetHudText('(%d, %d) %.4f' %
                            (x, y, self.raw_points['z'][y, x]))

    def Invalidate(self, globject=True, color=True):
        if not self.initialized:
            # haven't been initialized yet, no need to do any processing. It
            # will be done during the initialization.
            return

        if color:
            self.Colorize()
        if globject:
            # the objects will be re-generated before drawing
            self.globjects = None
        self.Refresh()

    def SetSelected(self, sel):
        self.selected.update(sel)
        self.UpdateHudText()
        self.Invalidate()

    def GetSelected(self):
        return self.selected

    def SetColorMap(self, clrmap):
        self.color_map = clrmap
        self.Invalidate()

    def GetColorMap(self):
        return self.color_map

    def SetColorScale(self, scale):
        self.color_scale = scale

    def GetColorScale(self):
        return self.color_scale

    def UpdateImage(self, points):
        # light-weighted function to update the image value as fast as possible,
        #assume its dimension does not change

        self.raw_points['z'] = points
        self.UpdateHudText()
        rows, cols = points.shape
        self.Pz[0:rows, 0:cols] = points
        self.Pz[-1, :] = self.Pz[-2, :]
        self.Pz[:, -1] = self.Pz[:, -2]
        if self.mode_2d and not self.show['surface']:
            # in 2d mode, point will only affect surface
            self.points[:, 2].fill(0)
        else:
            self.points[:, 2] = self.Pz.T.flatten()

        self.Invalidate()

    def SetDimension(self, rows, cols):
        # 'x', 'y' is used in 3D mode, where the image is expanded 1 pixel on
        # each dimension
        self.dimension = {
            'y': rows + 1,
            'x': cols + 1,
            'rows': rows,
            'cols': cols
        }
        # resize the horz/vert bar buffer
        self.horzbar_buf.Resize(cols)
        self.vertbar_buf.Resize(rows)
        # resize the image buffer
        self.Pz = np.zeros((rows + 1, cols + 1))
        self.points = np.zeros((self.Pz.size, 3))

    def GetDimension(self):
        return self.dimension

    def SetImage(self, points):
        # points is a dictionary with keys ['x', 'y', 'z'], and each should
        # be a 2D matrix with same dimension. 'z' is mandatory.
        # This function is slow, since it will allocate the memory and
        # regenerate x/y data if necessary
        self.raw_points = points
        rows, cols = points['z'].shape
        self.SetDimension(rows, cols)
        # expand the image by 1 pixel on each dimension, since for pixel (i, j)
        # the surface is draw as a quad defined by
        # (x(i) y(i), z(i)), (x(i+1) y(i), z(i))
        # (x(i) y(i+1), z(i)), (x(i+1) y(i+1), z(i))
        xy = np.meshgrid(np.arange(0, cols + 1), np.arange(0, rows + 1))
        # if 'x' data is not defined, assume it is [0:cols]
        if 'x' not in points:
            Px = xy[0]
        else:
            Px = np.zeros((rows + 1, cols + 1))
            Px[0:rows, 0:cols] = points['x']
            Px[-1, :] = Px[-2, :]
            if cols > 1:
                Px[:, -1] = Px[:, -2] * 2 - Px[:, -3]
            else:
                Px[:, -1] = Px[:, -2] + 1

        # if 'x' data is not defined, assume it is [0:cols]
        if 'y' not in points:
            Py = xy[1]
        else:
            Py = np.zeros((rows + 1, cols + 1))
            Py[0:rows, 0:cols] = points['y']
            Py[:, -1] = Py[:, -2]
            if rows > 1:
                Py[-1, :] = Py[-2, :] * 2 - Py[-3, :]
            else:
                Py[-1, :] = Py[-2, :] + 1

        self.Pz[0:rows, 0:cols] = points['z']
        self.Pz[-1, :] = self.Pz[-2, :]
        self.Pz[:, -1] = self.Pz[:, -2]
        xmin, ymin, zmin = np.min(Px), np.min(Py), np.min(self.Pz)
        xmax, ymax, zmax = np.max(Px), np.max(Py), np.max(self.Pz)
        if self.zrange:
            zmin, zmax = self.zrange
        self.data_offset['x'] = (xmax + xmin) / 2
        self.data_offset['y'] = (ymax + ymin) / 2
        self.data_offset['z'] = (zmax + zmin) / 2

        self.points = np.zeros((self.Pz.size, 3))
        self.points[:, 0] = Px.T.flatten()
        self.points[:, 1] = Py.T.flatten()
        self.points[:, 2] = self.Pz.T.flatten()
        self.SetRange({
            'xmin': xmin,
            'ymin': ymin,
            'zmin': zmin,
            'xmax': xmax,
            'ymax': ymax,
            'zmax': zmax
        })

        self.UpdateHudText()
        self.Invalidate()

    def CalcContour(self, v, level):
        # calculate the contour of level within a triangle defined by v
        # sort by Z
        v = v[v[:, 2].argsort()]

        p1 = None
        p2 = None
        valid = False
        if v[2][2] < level or v[0][2] > level:
            # do nothing
            pass
        elif v[0][2] < level and v[1][2] > level:
            p1 = (v[0] - v[1]) / (v[0][2] - v[1][2]) * (level - v[0][2]) + v[0]
            p2 = (v[0] - v[2]) / (v[0][2] - v[2][2]) * (level - v[0][2]) + v[0]
            valid = True
        elif v[1][2] < level and v[2][2] > level:
            p1 = (v[2] - v[0]) / (v[2][2] - v[0][2]) * (level - v[2][2]) + v[2]
            p2 = (v[2] - v[1]) / (v[2][2] - v[1][2]) * (level - v[2][2]) + v[2]
            valid = True
        elif v[0][2] == level and v[1][2] == level:
            p1 = v[0]
            p2 = v[1]
            valid = True
        elif v[1][2] == level and v[2][2] == level:
            p1 = v[1]
            p2 = v[2]
            valid = True

        return {'valid': valid, 'p1': p1, 'p2': p2}

    def PrepareContour(self, L, R):
        xmin, xmax = self.range['xmin'], self.range['xmax']
        ymin, ymax = self.range['ymin'], self.range['ymax']
        zmin, zmax = self.range['zmin'], self.range['zmax']
        # ugly, offset the contour to make it 'align' with mesh and quad
        ox = (xmax - xmin) / (self.dimension['x'] - 1) / 2
        oy = (ymax - ymin) / (self.dimension['y'] - 1) / 2
        oz = 0
        # calculate the contour levels uniformly, ignore endpoints (zmin, zmax)
        levels = np.linspace(zmin,
                             zmax,
                             self.contour_levels + 1,
                             endpoint=False)[1:]
        my = self.dimension['y']
        P = self.points
        vertexAll = []
        meshAll = []
        colourAll = []

        vertex = []
        mesh = []
        colour = []
        minmax = []
        for i in range(L, R):
            if i % my == my - 1 or i + my + 1 >= len(P):
                minmax.append([0, 0])
                continue
            t1 = min(P[i][2], P[i + 1][2], P[i + my][2], P[i + my + 1][2])
            t2 = max(P[i][2], P[i + 1][2], P[i + my][2], P[i + my + 1][2])
            minmax.append([t1, t2])

        for level in levels:
            clr = self.GetColorByZ(level)
            clr = [clr[0], clr[1], clr[2], 1]
            for i in range(L, R):
                if i % my == my - 1 or i + my + 1 >= len(P):
                    # the contour is calculate with for points
                    # i   i+my
                    # i+1 i+my+1
                    # so ignore the bottom and right edges
                    continue
                if minmax[i - L][0] > level or minmax[i - L][1] < level:
                    # the current level is not cross ith point and its 3
                    # neighbours, do nothing
                    continue
                if len(vertex) > 2**16 - 10:
                    # too many vertex, add to separate block to avoid overflow
                    # when rendering
                    vertex = (np.array(vertex) + [ox, oy, oz]).flatten()
                    colour = np.array(colour).flatten()
                    vertexAll.append(vertex)
                    meshAll.append(mesh)
                    colourAll.append(colour)
                    vertex = []
                    mesh = []
                    colour = []
                v = np.array([P[i], P[i + 1], P[i + my + 1]])
                p = self.CalcContour(v, level)
                if p['valid']:
                    vertex.append(p['p1'])
                    mesh.append(len(vertex) - 1)
                    vertex.append(p['p2'])
                    mesh.append(len(vertex) - 1)
                    colour.append(clr)
                    colour.append(clr)

                v = np.array([P[i], P[i + my], P[i + my + 1]])
                p = self.CalcContour(v, level)
                if p['valid']:
                    vertex.append(p['p1'])
                    mesh.append(len(vertex) - 1)
                    vertex.append(p['p2'])
                    mesh.append(len(vertex) - 1)
                    colour.append(clr)
                    colour.append(clr)
        if vertex:
            vertex = (np.array(vertex) + [ox, oy, oz]).flatten()
            colour = np.array(colour).flatten()
            mesh = np.array(mesh).flatten()
            vertexAll.append(vertex)
            meshAll.append(mesh)
            colourAll.append(colour)
        return {'Vertices': vertexAll, 'Mesh': meshAll, 'Color': colourAll}

    def PrepareGLObjects(self):
        colorAll = []
        vertexAll = []
        triangleAll = []
        meshAll = []
        contourAll = []
        block = int(len(self.points) * 4 / (2**16) + 1)
        my = self.dimension['y']
        mx = self.dimension['x']
        # the index of each block
        blocks = np.linspace(0, len(self.points), block + 1, dtype=int)
        self.blocks = blocks
        r = blocks[0]
        l = 0
        P = self.points
        for k in range(1, len(blocks)):
            l, r = r, blocks[k]
            acolor = np.repeat(self.pointsClr[np.arange(
                l, min(r + my + 1, len(P)))],
                               4,
                               axis=0)
            # for each point (x[i], y[j], z[i,j]), construct a quad
            # when show_step_surface is true
            # (x[i],   y[i],   z[i,j])
            # (x[i+1], y[i],   z[i,j])
            # (x[i+1], y[i+1], z[i,j])
            # (x[i],   y[i+1], z[i,j])
            # otherwise
            # (x[i],   y[i],   z[i,j])
            # (x[i+1], y[i],   z[i,j])
            # (x[i+1], y[i+1], z[i,j])
            # (x[i],   y[i+1], z[i,j])

            avertex = np.zeros(((min(r + my + 1, len(P)) - l) * 4, 3))
            # the index
            idx = np.arange(l, min(r + my + 1, len(P)))
            idx = np.matrix(idx).T
            idx = np.repeat(idx, 4, axis=1)
            # the corresponding row & col
            cc, rr = np.divmod(idx, my)
            #cc = (idx/my).astype(int)
            # the row for each 4 points in the same quad
            rr = rr + [0, 1, 1, 0]
            rr = rr.flatten().clip(0, my - 1)
            # the col for each 4 points in the same quad
            cc = cc + [0, 0, 1, 1]
            cc = cc.flatten().clip(0, mx - 1)
            # update the vertex of each quad
            avertex[:, 0] = P[(cc * my + rr), 0]
            avertex[:, 1] = P[(cc * my + rr), 1]
            if self.show_step_surface:
                # x, y
                avertex[:, 0:-1] = P[(cc * my + rr), 0:-1]
                # the quad use the same z value of (i, j)
                avertex[:, 2] = P[idx.flatten(), 2]
            else:
                avertex[:, :] = P[(cc * my + rr), :]

            # index for quad and mesh
            idx = np.arange(0, r - l)
            idx = np.matrix(idx).T
            idx = np.repeat(idx, 4, axis=1)
            COLS, ROWS = np.divmod(idx, my)
            idx = idx * 4

            # quad
            triangle = np.zeros(((r - l) * 3, 4))
            # the current quad
            # triangle.append([4*i, 4*i+1, 4*i+2, 4*i+3])
            triangle[0:r - l, :] = idx + [0, 1, 2, 3]

            # connect to the quad right
            # triangle.append([4*i+3, 4*i+2, 4*(i+my)+1, 4*(i+my)])
            rr = ROWS  #+ [0, 0, 0, 0]
            #rr = rr.clip(0, my-1)
            cc = COLS + [0, 0, 1, 1]
            cc = cc.clip(0, mx - 1)
            triangle[r - l:2 * (r - l), :] = (cc * my + rr) * 4 + [3, 2, 1, 0]

            # connect to the quad below
            # triangle.append([4*i+1, 4*(i+1), 4*(i+1)+3, 4*(i)+2])
            rr = ROWS + [0, 1, 1, 0]
            rr = rr.clip(0, my - 1)
            cc = COLS  #+ [0, 0, 0, 0]
            #cc = cc.clip(0, mx-1)
            triangle[2 * (r - l):3 *
                     (r - l), :] = (cc * my + rr) * 4 + [1, 0, 3, 2]

            # mesh
            mesh = None
            if self.show['mesh']:
                mesh = np.zeros(((r - l) * 4, 4))
                # mesh in the same quad
                # mesh.append([4*i, 4*i+1, 4*i+1, 4*i+2, 4*i+2, 4*i+3, 4*i+3, 4*i])
                mesh[0 * (r - l):1 * r - l] = idx + [0, 1, 1, 2]
                mesh[1 * (r - l):2 * r - l] = idx + [2, 3, 3, 0]

                # connect to the quad right
                # mesh.append([4*i+3, 4*(i+my), 4*i+2, 4*(i+my)+1])
                rr = ROWS  #+ [0, 0, 0, 0]
                #rr = rr.clip(0, my-1)
                cc = COLS + [0, 1, 0, 1]
                cc = cc.clip(0, mx - 1)
                mesh[2 * (r - l):3 *
                     (r - l), :] = (cc * my + rr) * 4 + [3, 0, 2, 1]

                # connect to the quad below
                # mesh.append([4*i+1, 4*(i+1), 4*i+2, 4*(i+1)+3])
                rr = ROWS + [0, 1, 0, 1]
                rr = rr.clip(0, my - 1)
                cc = COLS  #+ [0, 0, 0, 0]
                #cc = cc.clip(0, mx-1)
                mesh[3 * (r - l):4 *
                     (r - l), :] = (cc * my + rr) * 4 + [1, 0, 2, 3]

                mesh = mesh.flatten()
            contour = None
            if self.show['contour']:
                contour = self.PrepareContour(l, r)
            avertex = avertex.flatten()
            acolor = acolor.flatten()
            triangle = triangle.flatten()
            vertexAll.append(avertex)
            colorAll.append(acolor)
            triangleAll.append(triangle)
            meshAll.append(mesh)
            contourAll.append(contour)
        return {
            'block': block,
            'Color': colorAll,
            'Vertices': vertexAll,
            'Trinagles': triangleAll,
            'Mesh': meshAll,
            'Contour': contourAll
        }

    def SetGLBuffer(self, v, c):
        # vertex buffer
        # Bind it as The Current Buffer'
        glBindBuffer(GL_ARRAY_BUFFER, self.glBufs[0])
        # Fill it With the Data
        glBufferData(GL_ARRAY_BUFFER, v.astype(np.float32), GL_STATIC_DRAW)
        glVertexAttribPointer(self.VertexPosition, 3, GL_FLOAT, GL_FALSE, 0,
                              None)

        # color buffer
        glBindBuffer(GL_ARRAY_BUFFER, self.glBufs[1])
        glBufferData(GL_ARRAY_BUFFER, c.astype(np.float32), GL_STATIC_DRAW)
        glVertexAttribPointer(self.VertexColor, 4, GL_FLOAT, GL_FALSE, 0, None)

    def DrawElement(self, t, v, m, o=0):
        glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, self.glBufs[2])
        glBufferData(GL_ELEMENT_ARRAY_BUFFER, v.astype(np.uint16),
                     GL_STATIC_DRAW)
        glUniform1i(glGetUniformLocation(self.ShaderProgram, "mesh"), m)
        glUniform1i(glGetUniformLocation(self.ShaderProgram, "VertexOriginal"),
                    o)
        glDrawElements(t, v.size, GL_UNSIGNED_SHORT, None)

    def DrawAxis(self):
        r = self.range
        gx = 32 / self.scale['base']
        p = [r['xmax'], r['ymax'], r['zmax']]
        vertex = np.array([
            p[0] - gx, p[1], p[2], p[0] - gx, p[1] - gx, p[2], p[0], p[1] - gx,
            p[2], p[0], p[1], p[2], p[0] - gx, p[1], p[2], p[0], p[1], p[2],
            p[0] - gx / 3, p[1] + gx / 5, p[2], p[0] - gx / 3, p[1] - gx / 5,
            p[2]
        ])
        self.DrawAxisHelp('x', vertex)
        p = [r['xmin'], r['ymax'], r['zmax']]
        vertex = np.array([
            p[0], p[1], p[2], p[0], p[1] - gx, p[2], p[0] + gx, p[1] - gx,
            p[2], p[0] + gx, p[1], p[2], p[0], p[1] - gx, p[2], p[0], p[1],
            p[2], p[0] + gx / 5, p[1] - gx / 3, p[2], p[0] - gx / 5,
            p[1] - gx / 3, p[2]
        ])
        self.DrawAxisHelp('y', vertex)
        p = [r['xmin'], r['ymin'], r['zmax']]
        vertex = np.array([
            p[0], p[1] + gx, p[2] - gx, p[0], p[1], p[2] - gx, p[0], p[1],
            p[2], p[0], p[1] + gx, p[2], p[0], p[1], p[2] - gx, p[0], p[1],
            p[2], p[0], p[1] + gx / 5, p[2] - gx / 3, p[0], p[1] - gx / 5,
            p[2] - gx / 3
        ])
        self.DrawAxisHelp('z', vertex)

    def DrawAxisHelp(self, letter, vertex):
        W, H = 32, 32
        bitmap = wx.Bitmap(W, H, 32)
        tdc = wx.MemoryDC()
        tdc.SelectObject(bitmap)
        gc = wx.GraphicsContext.Create(tdc)
        font = wx.Font(14, wx.FONTFAMILY_DEFAULT, wx.FONTSTYLE_NORMAL,
                       wx.FONTWEIGHT_NORMAL)
        gc.SetFont(font, wx.WHITE)
        tw, th = gc.GetTextExtent(letter)
        gc.DrawText(letter, (W - tw) / 2, (H - th) / 2)
        tdc.SelectObject(wx.NullBitmap)

        texture = self.glTextures[0]
        glPixelStorei(GL_UNPACK_ALIGNMENT, GL_TRUE)
        glBindTexture(GL_TEXTURE_2D, texture)
        mybuffer = np.zeros((W, H, 4), np.uint8)
        bitmap.CopyToBuffer(mybuffer, wx.BitmapBufferFormat_RGBA)

        glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA, W, H, 0, GL_RGBA,
                     GL_UNSIGNED_BYTE, mybuffer.flatten())
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR)
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER,
                        GL_LINEAR_MIPMAP_NEAREST)
        glGenerateMipmap(GL_TEXTURE_2D)

        glBindTexture(GL_TEXTURE_2D, 0)

        glActiveTexture(GL_TEXTURE0)
        glBindTexture(GL_TEXTURE_2D, texture)

        color = np.array([
            0, 1, 0, 1, 0, 0, 0, 1, 1, 0, 0, 1, 1, 1, 0, 1, 1, 1, 1, 1, 1, 1,
            1, 1, 1, 1, 1, 1, 1, 1, 1, 1
        ])
        triangle = np.array([0, 1, 2, 0, 2, 3])
        mesh = np.array([4, 5, 5, 6, 5, 7])
        glUniform1i(glGetUniformLocation(self.ShaderProgram, "uSampler"), 0)
        self.SetGLBuffer(vertex, color)
        self.DrawElement(GL_TRIANGLES, triangle, 3)
        self.DrawElement(GL_LINES, mesh, 0)

    def Draw3dImg(self):
        if self.show['surface'] or self.show['mesh'] or self.show['contour']:
            if not self.globjects:
                self.globjects = self.PrepareGLObjects()
            obj = self.globjects
            block = obj.get('block', 0)
            glLineWidth(0.5)
            for b in range(block):
                self.SetGLBuffer(obj['Vertices'][b], obj['Color'][b])
                clr_mesh = 0
                if self.show['surface']:
                    # if surface is on, show the mesh in blue
                    clr_mesh = 1
                    self.DrawElement(GL_QUADS, obj['Trinagles'][b], 0)
                if self.show['mesh']:
                    self.DrawElement(GL_LINES, obj['Mesh'][b], clr_mesh, 3)

                if self.show['contour']:
                    ctr = obj['Contour'][b]
                    for c in range(len(ctr['Vertices'])):
                        self.SetGLBuffer(ctr['Vertices'][c], ctr['Color'][c])
                        self.DrawElement(GL_LINES, ctr['Mesh'][c], 0)

    def Draw2dImg(self):
        # Set Transformation Matrices, no rotation allowed in 2d mode
        TransformMatrix = np.eye(3, dtype=np.float32).flatten('F')
        tmatrix = glGetUniformLocation(self.ShaderProgram,
                                       "TransformationMatrix")
        glUniformMatrix3fv(tmatrix, 1, GL_FALSE, TransformMatrix)

        W = self.dimension['x']
        H = self.dimension['y']
        #if self.img_texture is None:
        glPixelStorei(GL_UNPACK_ALIGNMENT, GL_TRUE)
        glBindTexture(GL_TEXTURE_2D, self.glTextures[2])
        # the pointsClr is col-wise (i.e., 1st col, 2nd col...)
        mybuffer = self.pointsClr.astype(np.float32).flatten()
        glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA, H, W, 0, GL_RGBA, GL_FLOAT,
                     mybuffer)
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST)
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST)
        glGenerateMipmap(GL_TEXTURE_2D)

        glBindTexture(GL_TEXTURE_2D, 0)

        glActiveTexture(GL_TEXTURE0)
        glBindTexture(GL_TEXTURE_2D, self.glTextures[2])

        xmax, xmin = self.range['xmax'], self.range['xmin']
        ymax, ymin = self.range['ymax'], self.range['ymin']
        zmax, zmin = self.range['zmax'], self.range['zmin']
        vertex = np.array([
            xmin, ymax, zmin, xmin, ymin, zmin, xmax, ymin, zmin, xmax, ymax,
            zmin
        ])
        # texture coordinate
        color = np.array([
            1 - 1. / H, 0, 0, 1, 0, 0, 0, 1, 0, 1 - 1. / W, 0, 1, 1 - 1. / H,
            1 - 1. / W, 0, 1
        ])
        triangle = np.array([0, 1, 2, 3])
        glUniform1i(glGetUniformLocation(self.ShaderProgram, "uSampler"), 0)
        self.SetGLBuffer(vertex, color)
        self.DrawElement(GL_QUADS, triangle, 3)

    def DrawBox(self):
        glLineWidth(2)
        if self.show['box']:
            r = self.range
            vertex = np.array([
                r['xmin'], r['ymin'], r['zmin'], r['xmin'], r['ymax'],
                r['zmin'], r['xmax'], r['ymax'], r['zmin'], r['xmax'],
                r['ymin'], r['zmin'], r['xmin'], r['ymin'], r['zmax'],
                r['xmin'], r['ymax'], r['zmax'], r['xmax'], r['ymax'],
                r['zmax'], r['xmax'], r['ymin'], r['zmax']
            ])
            color = np.array([
                1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
                1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1
            ])
            axis = np.array([
                0, 1, 1, 2, 2, 3, 3, 0, 4, 5, 5, 6, 6, 7, 7, 4, 0, 4, 1, 5, 2,
                6, 3, 7
            ])
            self.SetGLBuffer(vertex, color)
            self.DrawElement(GL_LINES, axis, 0)

        if self.show['axis']:
            self.DrawAxis()

    def DrawBar(self):
        glLineWidth(2.0)
        if self.show['horz_bar']:
            y = self.selected['y']
            if y < 0 or y >= self.raw_points['z'].shape[0]:
                # out of range
                return
            d = self.raw_points['z'][y, :] * self.data_zscale
            self.horzbar_buf.SetData(d)
            v, c, l = self.horzbar_buf.GetGLObject()
            self.SetGLBuffer(v, c)
            self.DrawElement(GL_LINE_STRIP, l, 0)

        if self.show['vert_bar']:
            x = self.selected['x']
            if x < 0 or x >= self.raw_points['z'].shape[1]:
                # out of range
                return
            d = self.raw_points['z'][:, x] * self.data_zscale
            self.vertbar_buf.SetData(d)
            v, c, l = self.vertbar_buf.GetGLObject()

            self.SetGLBuffer(v, c)
            self.DrawElement(GL_LINE_STRIP, l, 0)

    def Draw(self):
        glClearColor(*self.background_color)
        glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT)

        #Setup shader
        vertexResolution = glGetUniformLocation(self.ShaderProgram,
                                                "VertexResolution")
        glUniform3f(vertexResolution, self.W, self.H, self.W)

        vertexOffset = glGetUniformLocation(self.ShaderProgram, "VertexOffset")
        ox = self.offset['base']['x'] + self.offset['user']['x']
        oy = self.offset['base']['y'] + self.offset['user']['y']
        glUniform3f(vertexOffset, ox, oy, self.offset['base']['x'])

        vertexOffsetData = glGetUniformLocation(self.ShaderProgram,
                                                "VertexOffsetData")
        glUniform3f(vertexOffsetData, self.data_offset['x'],
                    self.data_offset['y'], self.data_offset['z'])

        #Set Transformation Matrices
        TransformMatrix = self.rotate_matrix.flatten('F')
        tmatrix = glGetUniformLocation(self.ShaderProgram,
                                       "TransformationMatrix")
        glUniformMatrix3fv(tmatrix, 1, GL_FALSE, TransformMatrix)

        vertexScaleGL = glGetUniformLocation(self.ShaderProgram,
                                             "VertexScaleGL")
        glUniform3f(vertexScaleGL, 1, -1, -1)
        vertexScaleData = glGetUniformLocation(self.ShaderProgram,
                                               "VertexScaleData")
        glUniform3f(vertexScaleData, 1, 1, self.data_zscale)

        vertexScale = glGetUniformLocation(self.ShaderProgram, "VertexScale")
        scaleAll = self.scale['base'] * self.scale['zoom']
        glUniform3f(vertexScale, scaleAll, scaleAll, 1)

        if self.mode_2d:
            self.Draw2dImg()
        else:
            self.Draw3dImg()
            self.DrawBox()
        self.DrawBar()
        self.DrawHud()

    def DrawHud(self):
        if not self._hudtext and not self.hudtext:
            return
        W, H = int(self.W), int(self.H)
        HudW, HudH = self._hudBuffer.shape[0:2]
        if self._hudtext != self.hudtext or HudW < W:
            self._hudtext = self.hudtext
            letter = self.hudtext
            tdc = wx.MemoryDC()
            HudW, HudH = W, 38
            bitmap = wx.Bitmap(HudW, HudH, depth=32)
            tdc.SelectObject(bitmap)
            gc = wx.GraphicsContext.Create(tdc)
            font = wx.Font(14, wx.FONTFAMILY_TELETYPE, wx.FONTSTYLE_NORMAL,
                           wx.FONTWEIGHT_NORMAL)
            gc.SetFont(font, wx.WHITE)
            gc.DrawText(letter, 5, 5)
            tdc.SelectObject(wx.NullBitmap)
            self._hudBuffer = np.zeros((HudW, HudH, 4), np.uint8)
            bitmap.CopyToBuffer(self._hudBuffer, wx.BitmapBufferFormat_RGBA)
        texture = self.glTextures[1]
        glPixelStorei(GL_UNPACK_ALIGNMENT, GL_TRUE)
        glBindTexture(GL_TEXTURE_2D, texture)

        glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA, HudW, HudH, 0, GL_RGBA,
                     GL_UNSIGNED_BYTE, self._hudBuffer)
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST)
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST)
        glGenerateMipmap(GL_TEXTURE_2D)

        glBindTexture(GL_TEXTURE_2D, 0)

        glActiveTexture(GL_TEXTURE0)
        glBindTexture(GL_TEXTURE_2D, texture)

        vertex = np.array([
            -1, 1, -1, -1, 1 - 2. * HudH / H, -1, -1 + 2. * HudW / W,
            1 - 2. * HudH / H, -1, -1 + 2. * HudW / W, 1, -1
        ])
        # texture coordinate
        color = np.array([0, 0, 0, 1, 0, 1, 0, 1, 1, 1, 0, 1, 1, 0, 0, 1])
        triangle = np.array([0, 1, 2, 3])
        glUniform1i(glGetUniformLocation(self.ShaderProgram, "uSampler"), 0)
        self.SetGLBuffer(vertex, color)
        self.DrawElement(GL_QUADS, triangle, 3, 1)

    def Rotate(self, x, y, z):
        if x == 0 and y == 0 and z == 0:
            return
        R = self.rotate_matrix
        if x != 0:
            Rx = self.GetRotateMatrixX(x)
            R = np.matmul(Rx, R)
        if y != 0:
            Ry = self.GetRotateMatrixY(y)
            R = np.matmul(Ry, R)
        if z != 0:
            Rz = self.GetRotateMatrixZ(z)
            R = np.matmul(Rz, R)
        self.rotation = {'x': 0, 'y': 0, 'z': 0}
        self.rotate_matrix = R
        self.Refresh()

    def SetRotateDelta(self, delta):
        self.rotate_delta = delta

    def GetRotateDelta(self):
        return self.rotate_delta

    def GetRotateMatrixX(self, sign):
        Rx = np.zeros((3, 3), np.float32)
        theta = sign * self.rotate_delta
        Rx[0][0] = 1
        Rx[1][1] = math.cos(theta)
        Rx[1][2] = -math.sin(theta)
        Rx[2][1] = math.sin(theta)
        Rx[2][2] = math.cos(theta)
        return Rx

    def GetRotateMatrixY(self, sign):
        Ry = np.zeros((3, 3), np.float32)
        theta = sign * self.rotate_delta
        Ry[0][0] = math.cos(theta)
        Ry[0][2] = math.sin(theta)
        Ry[1][1] = 1
        Ry[2][0] = -math.sin(theta)
        Ry[2][2] = math.cos(theta)
        return Ry

    def GetRotateMatrixZ(self, sign):
        Rz = np.zeros((3, 3), np.float32)
        theta = sign * self.rotate_delta
        Rz[0][0] = math.cos(theta)
        Rz[0][1] = -math.sin(theta)
        Rz[1][0] = math.sin(theta)
        Rz[1][1] = math.cos(theta)
        Rz[2][2] = 1
        return Rz


class SelectedPixelBuf(object):
    def __init__(self, sz, rng):
        self.idx = 0
        self.range = dict(rng)
        self.line_color = [1., 1., 1., 1.]
        self.buf_size = -1
        self.Resize(sz)

    def SetRange(self, rng):
        if self.range == rng:
            return
        self.range.update(rng)
        self.Resize(self.buf_size)

    def SetColor(self, clr):
        if self.line_color == clr:
            return
        self.line_color = clr[:]
        if self.buf_size > 0:
            self.color = np.repeat(np.array([clr]), self.buf_size,
                                   axis=0).flatten()

    def Resize(self, sz):
        self.buf_size = sz
        self.buf = np.zeros(sz)
        # make z value larger than zmax, so it shows on top of the surface.
        self.vertex = np.ones((sz, 3)) * self.range['zmax'] * 2.0
        self.color = np.repeat(np.array([self.line_color]), sz,
                               axis=0).flatten()
        self.line = np.arange(sz)
        self.idx = 0

    def GetGLObject(self, l, r, t, b, o, g):
        self.vertex[:, 0] = np.linspace(l, r, self.buf.size, endpoint=False)
        self.vertex[:, 1] = b - (np.roll(self.buf, -(self.idx + 1)) - o) * g
        return self.vertex.flatten(), self.color, self.line

    def SetSelectedPos(self, pos):
        # pos is within [-self.buf.size+1, 0]
        self.color.fill(1)
        pos = (self.buf.size - 1 + pos) % self.buf.size
        if pos >= 0 and pos < self.buf.size:
            self.color[pos * 4:(pos + 1) * 4] = [1, 0, 0, 1]

    def SetData(self, d, idx):
        self.idx = idx
        self.buf[idx] = d


class TrackingSurface(SurfaceBase):
    DISPLAY_ORIGINAL = 0
    DISPLAY_MAX = 1
    DISPLAY_MIN = 2
    DISPLAY_MINMAX = 3
    ID_SIM_CLEAR = wx.NewId()
    ID_DISP_ORIGINAL = wx.NewId()
    ID_DISP_MAX = wx.NewId()
    ID_DISP_MIN = wx.NewId()
    ID_DISP_MINMAX = wx.NewId()

    def __init__(self, parent, points=None, buf_len=256):
        SurfaceBase.__init__(self, parent, points)
        self.buf_len = buf_len
        self.frames = None
        self.frames_idx = 0
        self.display_mode = self.DISPLAY_ORIGINAL
        self.frame_display = None
        self.selected_buf = None

    def Initialize(self):
        self.SetBufLen(self.buf_len)
        super(TrackingSurface, self).Initialize()

    def GetAccelList(self):
        accel_tbl = super(TrackingSurface, self).GetAccelList()
        accel_tbl += [(wx.ACCEL_CTRL, ord('K'), self.ID_SIM_CLEAR),
                      (wx.ACCEL_SHIFT, ord('M'), self.ID_DISP_MAX),
                      (wx.ACCEL_SHIFT, ord('O'), self.ID_DISP_ORIGINAL),
                      (wx.ACCEL_SHIFT, ord('N'), self.ID_DISP_MIN),
                      (wx.ACCEL_SHIFT, ord('R'), self.ID_DISP_MINMAX)]
        return accel_tbl

    def SetBufLen(self, sz):
        self.buf_len = sz
        self.frames = None
        self.frame_display = None
        self.frames_idx = 0
        self.selected_buf = SelectedPixelBuf(self.buf_len, self.range)

    def GetBufLen(self):
        return self.buf_len

    def SetRange(self, rng):
        super(TrackingSurface, self).SetRange(rng)
        if self.selected_buf:
            self.selected_buf.SetRange(rng)

    def UpdateHudText(self):
        if self.frames is None or not self.frames.size:
            self.hudtext = ""
            return
        _, rows, cols = self.frames.shape
        r = self.selected['y']
        c = self.selected['x']
        if r >= 0 and c >= 0 and r < rows and c < cols:
            d = self.frames[:, r, c]
            m = np.min(d)
            M = np.max(d)
            mn = np.mean(d)
            std = np.std(d)
            d = self.raw_points['z'][r, c]
            self.SetHudText(
                '(%d, %d) %f min: %f max: %f mean: %.2f std: %.2f' %
                (c, r, d, m, M, mn, std))

    def GetFrame(self, num):
        # get the num th frame,
        #   0: the latest frame
        #  -1: the 2nd latest frame
        #  ...
        if self.frames is None:
            return None
        idx = (self.frames_idx - 1 + num) % self.buf_len
        return self.frames[idx, :, :]

    def SetCurrentFrame(self, num):
        # show the num th frame,
        #   0: the latest frame
        #  -1: the 2nd latest frame
        #  ...
        if self.frames is None:
            return
        self.UpdateImage(self.GetFrame(num))
        self.selected_buf.SetSelectedPos(num)

    def NewFrameArrive(self, frame, silent=True):
        rows, cols = frame.shape
        if self.frames is None or rows != self.frames.shape[
                1] or cols != self.frames.shape[2]:
            self.frames = np.zeros((self.buf_len, rows, cols))
            self.frames_idx = 0
            self.SetImage({'z': frame})

        # all the following code is time sensitive, needs to finish as soon
        # as possible
        if self.frame_display is not None:
            if self.display_mode == self.DISPLAY_MAX:
                frame = np.maximum(frame, self.frame_display)
            elif self.display_mode == self.DISPLAY_MIN:
                frame = np.minimum(frame, self.frame_display)
            elif self.display_mode == self.DISPLAY_MINMAX:
                frame = np.where(
                    np.absolute(frame) > np.absolute(self.frame_display),
                    frame, self.frame_display)
        self.frame_display = frame

        if self.frames.size:
            self.frames[self.frames_idx, :, :] = frame
            r, c = self.selected['y'], self.selected['x']
            if r >= 0 and c >= 0 and r < rows and c < cols:
                self.selected_buf.SetData(frame[r, c], self.frames_idx)
            else:
                self.SetSelected({'x': 0, 'y': 0})

        if not silent:
            # update the data
            self.UpdateImage(frame)

        self.frames_idx += 1
        self.frames_idx %= self.buf_len

    def Clear(self):
        self.frame_display = None
        self.frames = None

    def Draw(self):
        super(TrackingSurface, self).Draw()
        if not self.frames is None:
            self.DrawSelectedBuf()

    def DrawSelectedBuf(self):
        if self.frames is None:
            return
        xmax, xmin = self.range['xmax'], self.range['xmin']
        ymax, ymin = self.range['ymax'], self.range['ymin']
        if self.default_rotate in [90, 270]:
            o = self.data_offset['y'] - self.data_offset['x']
            xmax, xmin, ymax, ymin = ymax - o, ymin - o, xmax, xmin
        v, c, l = self.selected_buf.GetGLObject(
            xmin, xmax, ymin, ymax,
            self.data_offset['z'] - self.data_offset['y'] / self.data_zscale,
            self.data_zscale)
        if len(l) >= 2**16 - 10:
            # if the buffer length is larger, draw line with multiple sections
            for s in range(0, len(l), 2**16 - 10):
                e = min(s + 2**16 - 10 + 1, len(l))
                self.SetGLBuffer(v[s * 3:e * 3], c[s * 4:e * 4])
                self.DrawElement(GL_LINE_STRIP, l[s:e] - l[s], 0, 2)
        else:
            self.SetGLBuffer(v, c)
            self.DrawElement(GL_LINE_STRIP, l, 0, 2)

    def GetContextMenu(self):
        menu = super(TrackingSurface, self).GetContextMenu()
        if not menu:
            menu = wx.Menu()
        else:
            menu.AppendSeparator()
        display = wx.Menu()
        display.Append(self.ID_DISP_ORIGINAL, 'Original\tshift+o', '',
                       wx.ITEM_CHECK)
        display.Append(self.ID_DISP_MAX, 'Max\tshift+m', '', wx.ITEM_CHECK)
        display.Append(self.ID_DISP_MIN, 'Min\tshift+n', '', wx.ITEM_CHECK)
        display.Append(self.ID_DISP_MINMAX, 'Min-Max\tshift+r', '',
                       wx.ITEM_CHECK)
        menu.AppendSubMenu(display, 'Display')
        return menu

    def OnUpdateMenu(self, event):
        eid = event.GetId()
        if eid == self.ID_DISP_ORIGINAL:
            event.Check(self.display_mode == self.DISPLAY_ORIGINAL)
        elif eid == self.ID_DISP_MAX:
            event.Check(self.display_mode == self.DISPLAY_MAX)
        elif eid == self.ID_DISP_MIN:
            event.Check(self.display_mode == self.DISPLAY_MIN)
        elif eid == self.ID_DISP_MINMAX:
            event.Check(self.display_mode == self.DISPLAY_MINMAX)
        else:
            super(TrackingSurface, self).OnUpdateMenu(event)

    def OnProcessMenuEvent(self, event):
        eid = event.GetId()
        if eid == self.ID_SIM_CLEAR:
            self.Clear()
        elif eid == self.ID_DISP_ORIGINAL:
            self.display_mode = self.DISPLAY_ORIGINAL
        elif eid == self.ID_DISP_MAX:
            self.display_mode = self.DISPLAY_MAX
        elif eid == self.ID_DISP_MIN:
            self.display_mode = self.DISPLAY_MIN
        elif eid == self.ID_DISP_MINMAX:
            self.display_mode = self.DISPLAY_MINMAX
        else:
            super(TrackingSurface, self).OnProcessMenuEvent(event)



"""
The code is based on PythonTookit (http://pythontoolkit.sourceforge.net/)
"""
import time
import inspect  #for debugger frame inpsection
import sys  #for set_trace etc
import os.path  #for absolute filename conversions
import ctypes  #for pythonapi calls
from codeop import _maybe_compile, Compile
import traceback  #for formatting errors
import wx
import wx.py.dispatcher as dp
import six.moves._thread as thread  #for keyboard interrupt

help_msg = """
\"\"\"
#help | #h:
    - show this text.
#step | #s:
    - step to the next line.
#stepin | #si:
    - step into a new code block call.
#stepout | #so:
    - step out of the current code block.
#resume | #r:
    - resume running the code.
#setscope [L] | #ss [L]:
    - set the active scope level L (default: 0 for the main user namespace).
#line | #l:
    - print the current line of source (if available).
#end | #e:
    - quit the debugger.
#stop:
    - interrupt the execution
\"\"\"
"""


class PseudoEvent(object):
    """
    An object with the same interface as a threading.Event for the internal
    engine. This prevents the readline/readlines and debugger from blocking
    when waiting for user input.

    This calls the doyield function and sleeps until the event is set.
    """
    def __init__(self):
        self._set = False

    def set(self):
        self._set = True

    def clear(self):
        self._set = False

    def wait(self, timeout=None):
        if timeout is not None:
            raise NotImplementedError('Timeout not implemented')
        while self._set is False:  #and (self._eng._stop is False):
            wx.YieldIfNeeded()
            # send the EVT_UPDATE_UI events so the UI status has a chance to
            # update (e.g., menubar, toolbar)
            wx.EventLoop.GetActive().ProcessIdle()
            time.sleep(0.05)


class EngineDebugger(object):
    bpnum = 0

    def __init__(self):
        self.compiler = EngineCompiler()
        #debugger state
        self._paused = False  #is paused.
        self._can_stepin = False  #debugger is about to enter a new scope
        self._can_stepout = False  #debugger can be stepped out of a scope

        #debugger command flags
        self._resume = False  #debugging was resumed.
        self._end = False  #stop debugging, finish running code
        self._stop = False  #stop running code
        self._stepin = False  #debugger step in to scope
        self._stepout = False  #debugger step out of scope

        #event used to wake the trace function when paused
        self._resume_event = PseudoEvent()

        #user commands to execute when the debugger is paused
        #this is check in the trace function and executed in the active scope.
        self._cmd = None

        #debugger scopes:
        #keep track of the different scopes available for tools to query
        self._scopes = []  #list of scope (function) names
        self._frames = []  #list of scope frames
        self._active_scope = 0  #the current scope level used for exec/eval/commands
        self._paused_active_scope = 0
        #internal variable used to keep track of where wer started debugging
        self._bottom_frame = None

        #files to look for when tracing...
        self._fncache = {}  #absolute filename cache
        self._block_files = [
        ]  #list of filepaths not to trace in (or any further)

        #prevent trace in all engine module files to avoid locks blocking comms threads.
        self.set_block_file(os.path.dirname(__file__) + os.sep + '*.*')

        # break points - info on breakpoints is stored in an DictList instance
        # which allows items (the breakpoint data dict) to be retrieved by
        #filename or id. The key tcount is the breakpoint counters for hits,
        # and is reset before each user command.
        self.bpoints = DictList()
        self._bp_hcount = {}  # hit counter {id: hcount}
        self._bp_tcount = {}  # trigger counter {id: tcount}

        #Register message handlers for the debugger
        dp.connect(self.pause, 'debugger.pause')
        dp.connect(self.resume, 'debugger.resume')
        dp.connect(self.stop_code, 'debugger.stop')
        dp.connect(self.end, 'debugger.end')
        dp.connect(self.step, 'debugger.step')
        dp.connect(self.step_in, 'debugger.step_into')
        dp.connect(self.step_out, 'debugger.step_out')
        dp.connect(self.set_scope, 'debugger.set_scope')
        dp.connect(self.set_breakpoint, 'debugger.set_breakpoint')
        dp.connect(self.clear_breakpoint, 'debugger.clear_breakpoint')
        dp.connect(self.edit_breakpoint, 'debugger.edit_breakpoint')
        dp.connect(self.get_breakpoint, 'debugger.get_breakpoint')
        dp.connect(self.get_status, 'debugger.get_status')

    def release(self):
        self.compiler.release()
        dp.disconnect(self.pause, 'debugger.pause')
        dp.disconnect(self.resume, 'debugger.resume')
        dp.disconnect(self.stop_code, 'debugger.stop')
        dp.disconnect(self.end, 'debugger.end')
        dp.disconnect(self.step, 'debugger.step')
        dp.disconnect(self.step_in, 'debugger.step_into')
        dp.disconnect(self.step_out, 'debugger.step_out')
        dp.disconnect(self.set_scope, 'debugger.set_scope')
        dp.disconnect(self.set_breakpoint, 'debugger.set_breakpoint')
        dp.disconnect(self.clear_breakpoint, 'debugger.clear_breakpoint')
        dp.disconnect(self.edit_breakpoint, 'debugger.edit_breakpoint')
        dp.disconnect(self.get_breakpoint, 'debugger.get_breakpoint')
        dp.disconnect(self.get_status, 'debugger.get_status')

    # Interface methods
    def reset(self):
        """Reset the debugger internal state"""
        self._paused = False
        self._can_stepin = False
        self._can_stepout = False

        self._resume = False
        self._end = False
        self._stop = False
        self._stepin = False
        self._stepout = False

        self._resume_event.clear()

        self._scopes = []
        self._frames = []
        self._active_scope = 0

        self._bottom_frame = None

        #resest breakpoint counters
        for d in self.bpoints.items():
            d['tcount'] = 0

    def stop_code(self):
        """
        Attempt to stop the running code by raising a keyboard interrupt in
        the main thread
        """
        self._stop = True  #stops the code if paused
        if self._paused is True:
            self.prompt()

        self._resume_event.set()

        #try a keyboard interrupt - this will not work for the internal engine
        # as the error is raised here instead of the running code, hence put in
        #try clause.
        try:
            thread.interrupt_main()
        except:
            pass

        # send the notification, the debugger may stop after the notification
        dp.send('debugger.ended')

    def end(self):
        """ End the debugging and finish running the code """
        #set the end flag and wake the traceback function
        #if necessary
        self._end = True
        self._resume = True
        self._resume_event.set()

        # send the notification, the debugger may stop after the notification
        dp.send('debugger.ended')
        return True

    def pause(self, flag=True):
        """ Pause currently running code - returns success """
        if flag is False:
            return not self.resume()

        #already paused do nothing.
        if self._paused is True:
            return True

        #set the paused flag to pause at next oppertunity
        self._paused = True
        return True

    def resume(self):
        """ Unpause currently running code """
        #not paused - so cannot resume
        if self._paused is False:
            return False
        #set flag to false
        self._resume = True
        #set resume event to wake the traceback function
        self._resume_event.set()
        return True

    def step(self):
        """
        Step to the next line of code when the debugger is paused
        """
        #not paused - cannot step
        if self._paused is False:
            return False
        #make sure the step_in flag is False, so we don't step into a new scope
        #but rather step 'over' it.
        self._stepin = False
        #set the event so the trace function can resume - but don't change the
        #paused flag so the code will pause again after the next line.
        self._resume_event.set()
        return True

    def step_in(self):
        """
        Step into the new scope (function/callable).
        This is only available if the next line is a 'call' event
        """
        #not paused or not at a call event
        if (self._paused is False) or (self._can_stepin) is False:
            return False

        #set the step in flag and wake the traceback function
        self._stepin = True
        self._resume_event.set()
        return True

    def step_out(self):
        """
        Step out of a scope (function/callable).
        This is only available if the debugger is currently in a scope other
        than the main user namespace.
        """
        #check if we can step out
        if (self._paused is False) or (self._can_stepout is False):
            return False

        #set the step out flag and wake the traceback function
        #this will turn the traceback off for the current scope by returning
        #None, as the new local trace function.
        self._can_stepout = False
        self._stepout = True
        self._resume_event.set()
        return True

    #breakpoints
    def set_breakpoint(self, bpdata):
        """
        Set a break point
        bpdata = {id, filename, lineno, condition, ignore_count, trigger_count}
        where id should be a unique identifier for this breakpoint
        """
        #check if the id to use already exists.
        if self.bpoints.filter(('id', ), (id, )):
            return False

        #check bpdata
        keys = list(bpdata.keys())
        if 'id' not in keys:
            bpdata['id'] = EngineDebugger.bpnum
            EngineDebugger.bpnum += 1
        elif 'filename' not in keys:
            return False
        elif 'lineno' not in keys:
            return False
        if 'condition' not in keys:
            bpdata['condition'] = ""
        if 'hitcount' not in keys:
            bpdata['hitcount'] = ""

        bpdata['tcount'] = 0

        #create new breakpoint
        bpdata['filename'] = self._abs_filename(bpdata['filename'])
        self.bpoints.append(bpdata)
        dp.send('debugger.breakpoint_added', bpdata=bpdata)
        return True

    def get_breakpoint(self, filename, lineno):
        """
        Return the break point according to the (filename, lineno),
        if not found, return None
        """
        bps = self.bpoints.filter(('filename', 'lineno'),
                                  (self._abs_filename(filename), lineno))
        if bps:
            return bps[0]
        return None

    def clear_breakpoint(self, id):
        """
        Clear the debugger breakpoint with the id given.
        If id is None all breakpoints will be cleared.
        """
        if id is None:
            self.bpoints.clear()
            return True

        #check if the id to clear exists.
        bps = self.bpoints.filter(('id', ), (id, ))
        if not bps:
            return False

        #remove the breakpoint
        bp = bps[0]
        self.bpoints.remove(bp)
        dp.send('debugger.breakpoint_cleared', bpdata=bp)
        return True

    def edit_breakpoint(self, id, **kwargs):
        """
        Edit a breakpoint.

        Only modify the keywords given (from filename, lineno, condition,
        trigger_count and ignore_count).

        e.g. edit_breakpoint(id=1, filename='test.py', lineno=23) will modify
        the breakpoint filename and lineno.
        """
        #check if the id to clear exists.
        bps = self.bpoints.filter(('id', ), (id, ))
        if not bps:
            return False

        bpdata = bps[0]
        if 'filename' in kwargs:
            kwargs['filename'] = self._abs_filename(kwargs['filename'])
        bpdata.update(kwargs)
        return True

    #debugger command/interogration interface
    def get_scope(self, level=None):
        """
        Get the scope name,frame, globals and locals dictionaries for the scope at the
        level given (None=current scope, 0=user namespace dictionary).

        Note: To make any changes permanent you should call
        _update_frame_locals(frame), with the frame.
        """
        #Note see http://utcc.utoronto.ca/~cks/space/blog/python/FLocalsAndTraceFunctions for possible problem!
        #Confirmed this issue - making a change to an existing variable will not
        #'stick' if the locals/globals are retrieved again before the next line
        #is executed.

        #To get round this either:
        #1)Keep a cache of the scopes globals/locals
        #to only fetch the dictionary once, after the next line the changes will
        #stick - but only to the top frame!
        #
        #2)Or using the python c api call that is called after the trace function
        #returns to make the changes stick immediately. This is done in the
        #function _update_frame_locals(frame)
        if level is None:
            level = self._active_scope
        if level > len(self._scopes) or level < 0:
            raise Exception('Level out of range: ' + str(level))
        #get the scope name
        name = self._scopes[level]
        frame = self._frames[level]
        return name, frame, frame.f_globals, frame.f_locals

    def set_scope(self, level, silent=False):
        """
        Set the scope level to use for interogration when paused. This will be
        reset at the next pause (i.e. after stepping).
        """
        #check level is an int
        if isinstance(level, int) is False:
            return False
        #check level is in range
        if level > (len(self._scopes) - 1):
            return False
        #check if already in this level
        if level == self._active_scope:
            return True
        self._active_scope = level

        #send scope change message
        if not silent:
            dp.send('debugger.update_scopes')
        return True

    def get_status(self):
        """return the current status"""
        if self._active_scope < 0 or self._active_scope >= len(self._frames):
            return None
        frame = self._frames[self._active_scope]
        filename = inspect.getsourcefile(frame) or inspect.getfile(frame)
        lineno = frame.f_lineno
        name = frame.f_code.co_name
        return {
            'name': name,
            'filename': self._abs_filename(filename),
            'lineno': lineno,
            'scopes': self._scopes,
            'active_scope': self._active_scope,
            'paused': self._paused,
            'can_stepin': self._can_stepin,
            'can_stepout': self._can_stepout,
            'frames': self._frames
        }

    def push_line(self, line):
        """
        A debugger commands to run - this is called from the engine when a
        line is pushed and we are debugging and paused, so store internally and
        wake the debugger.
        """
        #not paused so not expecting any commands???
        if self._paused is False:
            return
        #have some code to execute so call run_user_command to do something with it
        #this wakes the resume event and runs the code in the mainthread
        self._cmd = line
        self._resume_event.set()

    #filepaths
    def set_block_file(self, filepath):
        """
        Set a filepath to block from tracing.
        use path* to include multiple files.
        """
        filepath = self._abs_filename(filepath)
        self._block_files.append(filepath)

    def get_block_files(self):
        """
        Get the filepaths that are currently blocked from tracing
        """
        return self._block_files

    #engine like interfaces
    def evaluate(self, expression, level=None):
        """
        Evaluate expression in the active debugger scope and return result
        (like builtin eval)

        It is intened to be used to provide functionality to the GUI, so commands
        should be fairly quick to process to avoid blocking.

        The optional level argument controls the scope that will be used.
        scope=None uses the current debugger scope and scope=0 is the top level
        scope (the usernamespace).

        Returns None on errors
        """
        #get the scope locals/globals
        _, frame, g, l = self.get_scope(level)

        #try to evaluate the expression
        try:
            result = eval(expression, g, l)
        except:
            result = None

        #update the locals
        self._update_frame_locals(frame)
        return result

    def execute(self, expression, level=None):
        """
        Execute expression in engine (like builtin exec)

        It is intened to be used to provide functionality to the GUI interface,
        so commands should be fairly quick to process to avoid blocking the
        communications thread.

        The optional level argument controls the scope that will be used.
        scope=None uses the current debugger scope and scope=0 is the top level
        scope (the usernamespace).
        """
        #get the scope locals/globals
        _, frame, g, l = self.get_scope(level)

        #execute the expression
        try:
            exec(expression, g, l)
        except:
            pass

        #update the locals
        self._update_frame_locals(frame)

    # trace methods
    #Trace every line - why?
    #1) It allows step out to work from a breakpoint in a nested scope
    #2) It makes the code easier to follow
    #3) It allows running code to be paused anywhere rather than just in traced scopes
    #The downside is speed, with every line now causing a call into python code
    def __call__(self, frame, event, arg):
        """ This trace function is called only once when debugging starts."""
        #store the first frame so we know where user code starts
        self._bottom_frame = frame
        #set the new global trace function
        sys.settrace(self._trace_global)

        #update the scope list
        self._update_scopes(frame)

        #return the local trace function for the first call
        return self._trace_global(frame, event, arg)

    def _trace_global(self, frame, event, arg):
        """The main trace function called on call events """
        #Prepause
        #check if the engine wants to stop running code
        if self._stop is True:
            raise KeyboardInterrupt
        #check if the debugger want to end debugging.
        if self._end is True:
            sys.settrace(None)
            #use a blank trace function as returning None doesn't work
            return self._trace_off

        #file and name of scope being called.
        filename = inspect.getsourcefile(frame) or inspect.getfile(frame)
        lineno = frame.f_lineno

        #if the file is on the block list do not trace
        #this is mainly engine/message bus files and prevents the user pausing and
        #locking the engine communications.
        if self._check_files(filename, self._block_files):
            return None
        #if the calling frame is also blocked return None
        if frame.f_back.f_trace is None and frame != self._bottom_frame:
            return None

        #if trace_stepout is the parent frames trace function then do not pause
        #here unless there is a breakpoint set, just return the stepout trace function
        #which only checks for breakpoints and exits.
        elif (frame.f_back.f_trace) == (self._trace_stepout):
            return self._trace_stepout

        #check for breakpoints
        filename = self._abs_filename(filename)
        bps = self.bpoints.filter(('filename', 'lineno'), (filename, lineno))
        for bpdata in bps:
            paused = self._hit_bp(bpdata, frame)
            if paused is True:
                break
        #Pause
        ##pause at this line?
        if self._paused is True:
            #This pauses until stepped/resumed
            self._trace_pause(frame, event, arg)

        #After pause
        #if paused and stepping in print a message
        if self._paused:
            if self._stepin is True:
                local_trace = self._trace_local
            else:
                #not stepping in so use the stepout function
                local_trace = self._trace_stepout
        #not paused so carry on with the normal trace function
        else:
            local_trace = self._trace_local

        #reset the can step in flags
        self._can_stepin = False
        self._stepin = False

        return local_trace

    def _trace_local(self, frame, event, arg):
        """
        The local trace function
        - the main local trace function checks for breakpoints and pause requests
        - used for all traced scopes unless stepping out, or ending debugging.
        """
        #check if the engine wants to stop running code
        if self._stop is True:
            raise KeyboardInterrupt

        #check if the debugger want to end debugging.
        if self._end is True:
            sys.settrace(None)
            #use a blank trace function as returning None doesn't work
            #this is a python bug (as of python2.7 01/2011)
            return self._trace_off

        #file and name of scope being called.
        filename = inspect.getsourcefile(frame) or inspect.getfile(frame)
        lineno = frame.f_lineno

        #check for breakpoints
        filename = self._abs_filename(filename)
        bps = self.bpoints.filter(('filename', 'lineno'), (filename, lineno))
        for bpdata in bps:
            paused = self._hit_bp(bpdata, frame)
            if paused is True:
                break

        #Need to pause/already paused here
        if self._paused is True:
            #This pauses until stepped/resumed
            self._trace_pause(frame, event, arg)

        #check for step out
        if self._stepout is True:
            #set the previous frame to use the local trace function incase it is
            #not using it already
            frame.f_back.f_trace = self._trace_local
            #and use the stepout function from now on - _trace_local will only be
            #called again when a new breakpoint is encountered or if we are back
            #in the frame above (frame.f_back)
            self._stepout = False
            return self._trace_stepout

        return self._trace_local

    def _trace_stepout(self, frame, event, arg):
        """
        A minimial local trace function used when stepping out of a scope.
        - it will only pause if a breakpoint is encountered (it passes control back to trace_local)
        """
        #check if the engine wants to stop running code
        if self._stop is True:
            raise KeyboardInterrupt
        #check if the debugger want to end debugging.
        if self._end is True:
            sys.settrace(None)
            #use a blank trace function as returning None doesn't work
            return self._trace_off

        #file and name of scope being called.
        filename = inspect.getsourcefile(frame) or inspect.getfile(frame)
        lineno = frame.f_lineno

        #check for breakpoints
        filename = self._abs_filename(filename)
        bps = self.bpoints.filter(('filename', 'lineno'), (filename, lineno))
        for bpdata in bps:
            will_break = self._check_bp(bpdata, frame)
            if will_break is True:
                #do not bother to continue checking
                #pass to the full trace function.
                return self._trace_local(frame, event, arg)
        return self._trace_stepout

    def _trace_pause(self, frame, event, arg):
        """
        A function called from within the local trace
        function to handle a pauses at this event
        """
        #update the scope list
        self._update_scopes(frame)
        #if self._paused_active_scope != self._active_scope:
        #    return
        if event == 'call':
            self._can_stepout = False
            self._can_stepin = True
        elif len(self._scopes) > 1:
            self._can_stepout = True
            self._can_stepin = True
        else:
            self._can_stepout = False
            self._can_stepin = True
        if event == 'call' and not self._stepin:
            return

        # do not stop inside the system files
        filename = inspect.getsourcefile(frame) or inspect.getfile(frame)
        for f in sys.path + ['<input>', '<string>']:
            if filename.startswith(f):
                return

        #send a paused message to the console
        #(it will publish an ENGINE_DEBUG_PAUSED message after updating internal
        #state)
        dp.send('debugger.paused')
        #The user can then select whether to resume, step, step-in, step-out
        #cancel the code or stop debugging.

        #Make the console prompt for debugger commands at this pause.
        self.prompt()
        ##Paused loop - waiting for user instructions
        loop = True
        while loop:
            self._resume_event.wait()
            self._resume_event.clear()

            #check if the engine wants to stop running code
            if self._stop is True:
                raise KeyboardInterrupt

            #check if there is user code the execute
            if self._cmd is None:
                #turn console debug mode off -Note: None to turn off, True/False
                #is line continuation
                #need to step or resume so exit this while block
                #self.prompt()
                loop = False

            #user command to process.
            else:
                line = self._cmd
                self._cmd = None
                #check for debugger commands
                handled = self._process_dbg_command(line)
                if handled is False:
                    #not a command so execute as python source
                    self._process_dbg_source(line)

        #reset stepin flag
        self._can_stepin = False

        ##Debugger will run next line
        #check if we have resumed (will not pause again until a breakpoint/pause request)
        if self._resume is True:
            self._paused = False
            self._resume = False
            self._active_scope = 0

    def _trace_off(self, frame, event, arg):
        """
        A dummy tracing function used when ending tracing as returning None
        doesn't work (python bug v2.7 as of 01/2011)
        """
        return None

    # Internal methods
    def _check_bp(self, bpdata, frame):
        """
        Decide whether to break at the bpdata given.
        The frame is used to evaluate any conditons.
        """
        trigger = False
        #check if there is an expression to evaluate
        condition = bpdata['condition']
        if condition == "":
            hit = True
        else:
            try:
                hit = eval(condition, frame.f_globals, frame.f_locals)
            except:
                #fail safe - so trigger anyway.
                traceback.print_exc(file=sys.stdout)
                hit = True
                trigger = True

        if not trigger and hit:
            ht = bpdata['hitcount']
            if ht == "":
                trigger = True
            else:
                ht = ht.replace('#', str(bpdata['tcount']))
                try:
                    trigger = eval(ht, frame.f_globals, frame.f_locals)
                except:
                    #fail safe - so trigger anyway.
                    traceback.print_exc(file=sys.stdout)
                    trigger = True

        return trigger

    def _hit_bp(self, bpdata, frame):
        """
        Decide whether to break at the bpdata given and update internal counters
        and pause if necesary. The frame is used to evaluate any conditons.

        _hit_bp (this method) will check, update counters and pause if necessary
        _check_bp will only check - not pause or update counters.
        """
        trigger = False
        #checkif there is an expression to evaluate
        condition = bpdata['condition']
        if condition == "":
            #no condition triggering... increment tcount
            hit = True
        else:
            #evaluate it
            try:
                hit = eval(condition, frame.f_globals, frame.f_locals)
            except:
                #fail safe - so trigger anyway.
                traceback.print_exc(file=sys.stdout)
                hit = True
                trigger = True

        if not trigger and hit:
            bpdata['tcount'] += 1
            ht = bpdata['hitcount']
            if ht == "":
                trigger = True
            else:
                ht = ht.replace('#', str(bpdata['tcount']))
                try:
                    trigger = eval(ht, frame.f_globals, frame.f_locals)
                except:
                    #fail safe - so trigger anyway.
                    traceback.print_exc(file=sys.stdout)
                    trigger = True

        # pause if triggered
        if trigger is True:
            self._paused = True
            self._paused_active_scope = self._active_scope

        return trigger

    def _update_scopes(self, frame):
        scopes = []
        frames = []
        #loop backwards through frames until the bottom frame and generate lists
        #of scope names and frames.
        while frame is not None:
            name = frame.f_code.co_name
            if name == '<module>':
                filename = inspect.getsourcefile(frame) or inspect.getfile(
                    frame)
                if filename == '<input>':
                    name = 'Main'
            scopes.append(name)
            frames.append(frame)

            #check the next frame back
            if frame == self._bottom_frame:
                frame = None
            else:
                frame = frame.f_back

        #store to internals
        scopes.reverse()
        frames.reverse()
        self._scopes = scopes
        self._frames = frames

        #set current scope
        level = len(self._scopes) - 1
        if self._active_scope != level:
            self.set_scope(level, True)

    def _update_frame_locals(self, frame):
        """
        Ensures changes to a frames locals are stored so they are not lost.
        """
        # For explanation see get_scope and the following pages:
        # http://bugs.python.org/issue1654367#
        # http://www.gossamer-threads.com/lists/python/dev/546183
        #use PyFrame_LocalsToFast(f, 1) to save changes back into c array.
        func = ctypes.pythonapi.PyFrame_LocalsToFast
        func.restype = None
        func(ctypes.py_object(frame), 1)

    def _abs_filename(self, filename):
        """
        Internal method to return the absolute filepath form.
        """
        #change filename provided to absolute form
        #This is done using code from the bdb.BdB.canonic() method
        if filename == "<" + filename[1:-1] + ">":
            #file name is a special case - do not modify
            fname = filename
        else:
            #check in cache of filenames already done.
            fname = self._fncache.get(filename, None)
            if not fname:
                #not in cache, get the absolute path
                fname = os.path.abspath(filename)
                fname = os.path.normcase(fname)
                self._fncache[filename] = fname
        return fname

    def _check_files(self, filename, filelist):
        """
        Internal method to check if the file is included in the list (or included
        by a wildcard *).Returns True/False.
        """
        #get absolute filepath form
        fname = self._abs_filename(filename)

        #loop over filelist entries checking each one - return on first positive.
        for f in filelist:
            #a wildcard - check if fname startswith the path upto the *
            if f.endswith('*'):
                if fname.startswith(f[:-1]):
                    return True
            #a normal path - check if fname is this path.
            elif fname == f:
                return True
        return False

    def _process_dbg_command(self, line):
        """
        Check for, and perform user debugger commands.
        Returns handled=True/False
        """
        ##Check for debugger comment commands:
        if line.startswith('#') is False:
            return False

        cmd = line.rstrip('\n')  #remove the trailing newline for this check
        parts = cmd.split(' ')  #split into command and arguments
        cmd = parts[0]
        args = parts[1:]
        cmd = cmd.lower()  #convert to all lower case:

        #step
        if cmd in ['#step', '#s']:
            res = self.step()
            if res is False:
                self.write_debug('Cannot step here')
            else:
                #no need to prompt as stepping
                return True
        #step in
        elif cmd in ['#stepin', '#si']:
            res = self.step_in()
            if res is False:
                self.write_debug('Cannot step in here')
            else:
                #no need to prompt as stepping
                return True
        #stepout
        elif cmd in ['#stepout', '#so']:
            res = self.step_out()
            if res is False:
                self.write_debug('Cannot step out here')
            else:
                #no need to prompt as stepping
                return True

        #set scope
        elif cmd in ['#setscope', '#ss']:
            if len(args) != 1:
                level = 0
            try:
                level = int(args[0])
            except:
                level = args[0]
            res = self.set_scope(level)
            if res is False:
                msg = 'Usage: #setscope level (0 to %d)' % (len(self._scopes) -
                                                            1)
                self.write_debug(msg)
        #resume
        elif cmd in ['#resume', '#r']:
            res = self.resume()
            if res is False:
                self.write_debug('Cannot resume here')
            else:
                #no need to prompt as resuming
                return True
            return True
        #end debugging
        elif cmd in ['#end', '#e']:
            self.end()
            return True
        #help
        elif cmd in ['#help', '#h']:
            self.write_debug(help_msg)

        elif cmd in ['#line', '#l']:
            frame = self._frames[self._active_scope]
            tb = inspect.getframeinfo(frame)
            if tb.code_context is None:
                self.write_debug('Source line unavailable')
            else:
                self.write_debug(tb.code_context[tb.index])
        elif cmd in ['#stop']:
            self.stop_code()
            return True
        else:
            return False

        #Prompt for a new command and return
        self.prompt()
        return True

    def _process_dbg_source(self, line):
        """
        Process a line of user input as python source to execute in the active
        scope
        """
        ##line is user python command compile it
        ismore, code, err = self.compiler.compile(line)

        #need more
        #   - tell the console to prompt for more
        if ismore:
            self.prompt(ismore=True)
            return

        #syntax error
        #   - compiler will output the error
        #   - tell the console to prompt for new command
        if err:
            self.prompt(iserr=True)
            return

        #no code object - could be a blank line
        if code is None:
            self.prompt()
            return

        ##run the code in the active scope
        _, frame, g, l = self.get_scope(self._active_scope)
        try:
            exec(code, g, l)
        except SystemExit:
            self.write_debug('Blocking system exit')
        except KeyboardInterrupt:
            self._paused = False
            raise KeyboardInterrupt
        except:
            #engine wanted to stop anyway - probably wxPython keyboard interrupt error
            if self._stop is True:
                self._paused = False
                raise KeyboardInterrupt

            #error in user code.
            self.compiler.show_traceback()

        #update the locals
        self._update_frame_locals(frame)

        ##Finished running the code - prompt for a new command
        # add the command to history
        dp.send('shell.add_to_history', command=line)
        self.prompt()

    def write_debug(self, string):
        """
        Write a debugger message to the controlling console
        """
        print(string)

    def prompt(self, ismore=False, iserr=False):
        dp.send('shell.prompt', ismore=ismore, iserr=iserr)


"""
Engine misc.

Various engine utility functions/classes
"""


# A general purpose list type object used for storing dictionaries these can
# then be filtered by key.
# Used to store breakpoints in both the engine debugger (engine process) and
# engine manager(gui process) where breakpoints are stored as dictionaries
class DictList(object):
    def __init__(self, dicts=None, default=None):
        """
        Create a list like container object for dictionaries with the ability to
        look up dicts by key value or by index.

        dicts - is a sequence of dictionaries to populate the DictList with.
        default - a defualt dictionary to use to create a new dictionary in the
        list.

        Example.

        >>> person1 = {'Name':'John', 'Age':53}
        >>> person2 = {'Name':'Bob', 'Age':53}
        >>> person3 = {'Name':'Sally', 'Age':31}
        >>> default = {'Name':'', 'Age':0}
        >>> dlist = DictList((person1,person2,person3),default)

        >>> #Filter by age
        >>> dlist.filter(keys=('Age',),values=(53,))
        [{'Age': 53, 'Name': 'John'}, {'Age': 53, 'Name': 'Bob'}]

        >>> #Filter by age and name
        >>> dlist.filter(keys=('Age','Name'),values=(53,'John'))
        [{'Age': 53, 'Name': 'John'}]

        >>> #Add a new dictionary
        >>> dlist.new()
        {'Age': 0, 'Name': ''}
        """
        if dicts is None:
            dicts = []
        if default is None:
            default = {}
        self._dicts = []
        for d in dicts:
            self._dicts.append(d)
        self._default = default

    def clear(self):
        """Remove all items"""
        self._dicts = []

    def index(self, d):
        """Find the index of a dictionary in the list"""
        return self._dicts.index(d)

    def pop(self, index):
        """Remove the dictionary at the index given"""
        return self._dicts.pop(index)

    def items(self, key=None):
        """
        Return a list of all dictionaries in the list. If the optional key is
        given it will returns a list of the values of all dicts which have that
        key.
        """
        if key is None:
            return self._dicts

        #get the key values
        dicts = []
        for d in self._dicts:
            if key in d:
                dicts.append(d[key])
        return dicts

    def filter(self, keys=(), values=()):
        """
        Filter the dictionaries in the list to return only those
        with all matching key value pairs.
        """
        dicts = self._dicts
        #filter for each key,value in
        for key, value in zip(keys, values):
            self._fkey = key
            self._fvalue = value
            dicts = list(filter(self._filter, dicts))
            self._fkey = None
            self._fvalue = None
        return dicts

    def values(self, key):
        """
        Return a list of all values a given key has in the dictionarys.
        """
        values = []
        for i in self.items():
            value = i[key]
            if value not in values:
                values.append(value)
        return values

    def append(self, d):
        """
        Append a dictionary to the list.
        """
        #check item
        if isinstance(d, dict) is False:
            raise ValueError('Expected a dictionary')
        self._dicts.append(d)

    def new(self):
        """
        Create and append a new dictionary to the list. If a default dictionary
        has been set a shallow copy will made. The new dictionary is returned.
        """
        d = self._default.copy()
        self.append(d)
        return d

    def remove(self, d):
        """
        Remove the dictionary from the list and return it.
        """
        n = self.index(d)
        return self.pop(n)

    def set_default(self, d):
        if isinstance(d, dict) is False:
            raise ValueError('Default should be a dictionary')

    def _filter(self, d):
        """internal method to find matching items"""
        #d is the dictionary to check for a match
        #self._fkey, self._fvalue are stored before calling this function and
        #are the key:value to match
        if self._fkey in d:
            return d[self._fkey] == self._fvalue
        return False

    def __len__(self):
        return len(self._dicts)

    def __iter__(self):
        return self._dicts.__iter__()

    def __getitem__(self, index):
        return self._dicts[index]

    def __setitem__(self, index, d):
        #index store new item
        if isinstance(d, dict) is False:
            raise ValueError('Expected a dictionary')
        self._dicts[index] = d

    def __repr__(self):
        s = 'DictList(' + self._dicts.__repr__() + ')'
        return s


"""
Engine Compiler

Handles the compilation of source and formating exceptions and tracebacks for
the engine.
"""


class EngineCompiler(Compile):
    def __init__(self):
        Compile.__init__(self)

        #the number of lines to remove from the traceback line count,
        #for multi line hack in compilation stage
        self._lineadjust = 0

        #source buffer to get print line of source in exception
        self._buffer = ''

        #register message handlers
        dp.connect(self.future_flag, 'debugger.futureflag')

    def release(self):
        dp.disconnect(self.future_flag, 'debugger.futureflag')

    # Interface methods
    def set_compiler_flag(self, flag, is_set=True):
        """
        Set or unset a __future__ compiler flag
        """
        if is_set:
            self.flags |= flag
        else:
            self.flags &= ~flag

    def compile(self, source):
        """
        Compile source.
        Returns (more,code,err) where:

        more        -   Bool indicating whether more is expected
        code        -   the compiled code object if any
        err         -   True if syntax error, false otherwise
        """
        ##to enable more lines when the current code is ok but not finished we
        ##remove the final \n, this means the line must have \n\n to be complete.
        #if source[-1]=='\n':
        #    source = source[:-1]
        # Check for source consisting of only blank lines and comments
        for line in source.split("\n"):
            line = line.strip()
            if line and line[0] != '#':
                break  # Leave it alone
        else:
            source = ""
        ##check that the line is not empty
        if source.lstrip() == '':
            return False, None, False

        ##to enable multiple lines to be compiled at once we use a cheat...
        ##everything is enclosed in if True: block so that it will compile as a
        ##single line, but need to check that there is no indentation error...
        if source[0] != ' ':  #check for first line indentation error
            self._lineadjust = -1
            lastline = source.split('\n')[-1]
            source = "if True:\n " + source.replace('\n', '\n ')
            #the old source ended with a '\n' no more to come so add '\n'
            #so that it compiles
            if source.endswith('\n '):
                source = source + '\n'
            #also if the last line is zero indent add a \n as the block is also
            #complete
            if (len(lastline) - len(lastline.lstrip())) == 0:
                source = source + '\n'
        else:
            self._lineadjust = 0

        #store source in buffer
        self._buffer = source

        ##now compile as usual.
        filename = "<Engine input>"
        symbol = "single"
        try:
            code = _maybe_compile(self, source, filename, symbol)
        except (OverflowError, SyntaxError, ValueError):
            # Case 1 - syntax error
            self.show_syntaxerror()
            return False, None, True

        if code is None:
            # Case 2 - more needed
            return True, None, False

        # Case 3 - compile to code
        return False, code, False

    def show_syntaxerror(self):
        """
        Display the syntax error that just occurred.
        This doesn't display a stack trace because there isn't one.

        If a filename is given, it is stuffed in the exception instead
        of what was there before (because Python's parser always uses
        "<string>" when reading from a string).

        taken from: code.InteractiveInterpreter
        """
        filename = "<Engine input>"

        errtype, value, sys.last_traceback = sys.exc_info()
        sys.last_type = errtype
        sys.last_value = value

        try:
            msg, (dummy_filename, lineno, offset, line) = value
        except:
            # Not the format we expect; leave it alone
            pass
        else:
            if dummy_filename == filename:
                #(lineno-1 due to multiline hack in compile source)
                value = errtype(
                    msg, (filename, lineno + self._lineadjust, offset, line))
            else:
                value = errtype(msg, (dummy_filename, lineno, offset, line))
            sys.last_value = value
        err = traceback.format_exception_only(errtype, value)
        map(sys.stderr.write, err)

    def show_traceback(self):
        """
        Display the exception that just occurred.
        We remove the first stack item because it is our own code and adjust the
        line numbers for any engine inputs.
        modified from: code.InteractiveInterpreter
        """
        try:
            typ, value, sys.last_traceback = sys.exc_info()
            sys.last_type = typ
            sys.last_value = value
            tblist = traceback.extract_tb(sys.last_traceback)
            del tblist[:1]  #in our code so remove

            for n in range(0, len(tblist)):
                filename, lineno, offset, line = tblist[n]
                if filename == "<Engine input>":
                    #alter line number
                    tblist[n] = (filename, lineno + self._lineadjust, offset,
                                 line)
                lst = traceback.format_list(tblist)
                if lst:
                    lst.insert(0, "Traceback (most recent call last):\n")
                lst[len(lst):] = traceback.format_exception_only(typ, value)
            map(sys.stderr.write, lst)
        except:
            pass

    # Message handlers
    def future_flag(self, msg):
        """enable or disable a __future__ feature using flags"""
        flag, is_set = msg.get_data()
        self.set_compiler_flag(flag, is_set)



import sys
import os
import traceback
import inspect
import re
import ctypes
import six.moves.queue as Queue
import six
import numpy as np
from . import csim

SC_OBJ_UNKNOWN = 0
SC_OBJ_SIGNAL = 1
SC_OBJ_INPUT = 2
SC_OBJ_OUTPUT = 3
SC_OBJ_INOUT = 4
SC_OBJ_CLOCK = 5
SC_OBJ_XSC_PROP = 6
SC_OBJ_XSC_ARRAY_ITEM = 7
SC_OBJ_MODULE = 8
SC_OBJ_XSC_ARRAY = 9

BSM_FS = 0
BSM_PS = 1
BSM_NS = 2
BSM_US = 3
BSM_MS = 4
BSM_SEC = 5

BSM_POSEDGE = 0
BSM_NEGEDGE = 1
BSM_BOTHEDGE = 2
BSM_NONEEDGE = 3

BSM_TRACE_VCD = 0
BSM_TRACE_SIMPLE = 1

BSM_DATA_STRING = 0
BSM_DATA_FLOAT = 1
BSM_DATA_INT = 2
BSM_DATA_UINT = 3


class SimSysC(object):
    """class to load the systemc simulation"""
    def __init__(self, dll):
        self.ctx_callback = None
        self.valid = False
        try:
            folder = os.path.dirname(os.path.realpath(__file__))
            self.csim = csim.init_dll(dll, os.path.join(folder, 'bsm.h'))

            # create simulation
            ctx = self.csim.sim_context()
            self.csim.bsm_sim_top(ctx)
            self.ctx = csim.SStructWrapper(ctx)

            # load all objects
            obj = self.csim.sim_object()
            rtn = self.csim.ctx_first_object(obj)
            self.sim_objects = {}
            while rtn:
                wobj = csim.SStructWrapper(obj)
                self.check_object(wobj)
                if obj.readable:
                    self.csim.ctx_read(obj)
                self.sim_objects[wobj.name] = wobj
                obj = self.csim.sim_object()
                rtn = self.csim.ctx_next_object(obj)
            self.valid = True
        except:
            traceback.print_exc(file=sys.stdout)

    def __getattr__(self, item):
        # so we can call simulation functions 'directly'
        if item == 'csim':
            raise AttributeError()
        if 'ctx' in item:
            return getattr(self.csim, item)
        raise AttributeError()

    def is_valid(self):
        return self.valid

    def get_objects(self):
        if not self.is_valid():
            return {}
        return self.sim_objects

    def check_object(self, obj):
        if not obj or (not isinstance(obj, csim.SStructWrapper)):
            return

        kinds = {
            'sc_signal': SC_OBJ_SIGNAL,
            'sc_in': SC_OBJ_INPUT,
            'sc_out': SC_OBJ_OUTPUT,
            'sc_in_out': SC_OBJ_INOUT,
            'sc_clock': SC_OBJ_CLOCK,
            'xsc_property': SC_OBJ_XSC_PROP,
            'sc_module': SC_OBJ_MODULE,
            'xsc_array': SC_OBJ_XSC_ARRAY
        }
        kind = kinds.get(obj.kind, SC_OBJ_UNKNOWN)
        if kind == SC_OBJ_XSC_PROP and '[' in obj.name and ']' in obj.name:
            kind = SC_OBJ_XSC_ARRAY_ITEM
        obj.nkind = kind
        obj.register = (kind in (SC_OBJ_SIGNAL, SC_OBJ_INPUT, SC_OBJ_OUTPUT,
                                 SC_OBJ_INOUT, SC_OBJ_CLOCK, SC_OBJ_XSC_PROP,
                                 SC_OBJ_XSC_ARRAY_ITEM))

        # retrieve the parent module from name
        name = obj.name
        idx = name.rfind('.')
        if kind == SC_OBJ_XSC_ARRAY_ITEM:
            idx = name.rfind('[')
        obj.parent = ""
        if idx != -1:
            obj.parent = name[0:idx]

    def find_object(self, obj):
        if not self.is_valid():
            return None
        if isinstance(obj, six.string_types):
            obj = self.sim_objects.get(obj, None)
        if obj and isinstance(obj, csim.SStructWrapper) and \
           obj.is_type(self.csim.sim_object):
            return obj
        return None

    def ctx_read(self, obj):
        obj = self.find_object(obj)
        if obj is None:
            return ""

        obj = obj()
        if not obj.readable:
            return ""

        val = ''
        if self.csim.ctx_read(obj):
            if obj.value.type == BSM_DATA_STRING:
                val = ctypes.cast(obj.value.sValue,
                                  ctypes.c_char_p).value.decode()
            elif obj.value.type == BSM_DATA_FLOAT:
                val = obj.value.fValue
            elif obj.value.type == BSM_DATA_INT:
                val = obj.value.iValue
            elif obj.value.type == BSM_DATA_UINT:
                val = obj.value.uValue
        return val

    def ctx_write(self, obj, value):
        obj = self.find_object(obj)
        if obj is None:
            return False

        obj = obj()

        if not obj.writable:
            return False

        if obj.value.type == BSM_DATA_STRING:
            obj.value.sValue = (ctypes.c_byte * len(obj.value.sValue))(
                *bytearray(value))
        elif obj.value.type == BSM_DATA_FLOAT:
            obj.value.fValue = float(value)
        elif obj.value.type == BSM_DATA_INT:
            obj.value.iValue = int(value)
        elif obj.value.type == BSM_DATA_UINT:
            obj.value.uValue = int(value)
        else:
            return False

        return self.csim.ctx_write(obj)

    def ctx_time_str(self):
        if not self.is_valid():
            return ""
        buf = ctypes.create_string_buffer(256)
        self.csim.ctx_time_str(ctypes.cast(buf, ctypes.POINTER(ctypes.c_byte)))
        return ctypes.cast(buf, ctypes.c_char_p).value.decode()

    def ctx_set_callback(self, fun):
        # the callback fun should take an integer arguments and return an
        # integer argument. Return 1 from the callback function will pause
        # the simulation
        if not self.is_valid():
            return False
        self.ctx_callback = csim.callback(self.csim.bsm_callback(), fun)
        self.csim.ctx_set_callback(self.ctx_callback)
        return True


class BpCond(object):
    """
    The class to handle the breakpoint condition. One breakpoint may trigger on
    multiple conditions, and each condition may trigger on multiple hit-counts.
    """
    def __init__(self, condition, hitcount):
        # single condition
        self.condition = condition
        self.valid = True
        # multiple hitcounts on multiple conditions
        # since they trigger on the same condition, it only needs to evaluate
        # the condition once
        self.hitcount = []
        if isinstance(hitcount, list):
            self.hitcount = hitcount
        else:
            self.hitcount = [hitcount]
        # generate the hitcount set to remove the duplicated hitcounts, so
        # they are not evaluated multiple times
        self.hitcountset = set(self.hitcount)
        # the number of the breakpoint condition being triggered since it is set
        self.hitsofar = 0

    def add_hitcount(self, hitcount):
        # append the hitcount
        self.hitcount.append(hitcount)
        # regenerate the set
        self.hitcountset = set(self.hitcount)

    def del_hitcount(self, hitcount):
        idx = self.hitcount.index(hitcount)
        if idx == -1:
            return
        del self.hitcount[idx]
        self.hitcountset = set(self.hitcount)

    def triggered(self, val, valp):
        # invalid condition, no need to check
        if not self.valid:
            return None
        try:
            trigger = False
            # if the condition is empty, trigger if the value has changed
            if not self.condition:
                trigger = (val != valp)
            else:
                # replace the '$' with the current value, then evaluate it
                cond = self.condition.replace('$', str(val))
                trigger = eval(cond)

            if trigger:
                self.hitsofar += 1
                for hc in self.hitcountset:
                    # if the hitcount is empty, always trigger
                    if not hc:
                        return (hc, self.hitsofar)
                    # replace the '#' with the hitsofar
                    hce = hc.replace('#', str(self.hitsofar))
                    bk = eval(hce)
                    if bk is True:
                        return (hc, self.hitsofar)
        except:
            traceback.print_exc(file=sys.stdout)
            self.valid = False
        return None

    def __len__(self):
        # return the hitcount length, it may be called to determine whether
        # the condition is empty and can be deleted.
        return len(self.hitcount)


class Breakpoint(object):
    def __init__(self, name):
        self.name = name
        # the BpCond list
        self.condition = []

    def add_cond(self, cond, hitcount):
        for cnd in self.condition:
            # add hitcount to the existing condition
            if cnd.condition == cond:
                cnd.add_hitcount(hitcount)
                return
        # create a new BpCond instance
        self.condition.append(BpCond(cond, hitcount))

    def del_cond(self, cond, hitcount):
        for cnd in self.condition:
            if cnd.condition == cond:
                cnd.del_hitcount(hitcount)
                if len(cnd) <= 0:
                    del self.condition[self.condition.index(cnd)]
                break

    def triggered(self, val, valp):
        for cnd in self.condition:
            trigger = cnd.triggered(val, valp)
            if trigger is not None:
                return (cnd.condition, trigger[0], trigger[1])
        return None

    def __len__(self):
        # it may be called to determine whether the bp is empty and can be
        # deleted
        return len(self.condition)


class BpList(object):
    def __init__(self):
        self.data = {}
        self.data_raw = []

    def add(self, objs, objectsdict):
        resp = {}
        for name, cond, hitcount in objs:
            resp[name] = False
            if name not in six.iterkeys(objectsdict):
                continue
            if name in six.iterkeys(self.data):
                self.data[name].add_cond(cond, hitcount)
            else:
                self.data[name] = Breakpoint(name)
                self.data[name].add_cond(cond, hitcount)
            resp[name] = True
            self.data_raw.append([name, cond, hitcount])
        return resp

    def delete(self, objs, objectsdict):
        resp = {}
        for name, cond, hitcount in objs:
            resp[name] = False
            if name not in six.iterkeys(objectsdict):
                continue
            if name not in six.iterkeys(self.data):
                continue
            self.data[name].del_cond(cond, hitcount)
            if len(self.data[name]) <= 0:
                del self.data[name]
            resp[name] = True
            idx = self.data_raw.index([name, cond, hitcount])
            del self.data_raw[idx]
        return resp

    def get_bp(self):
        return self.data


# the list of registers to be monitored
class MonitorList(object):
    def __init__(self):
        self.data = {}
        self.monitor = []

    def add(self, objs, objsdict):
        resp = {o: False for o in objs}
        if objsdict:
            objs = [o for o in objs if o in six.iterkeys(objsdict)]
        for obj in objs:
            self.data[obj] = self.data.get(obj, 0) + 1
            resp[obj] = True
        self.update_monitor(objsdict)
        return resp

    def delete(self, objs, objsdict):
        resp = {o: False for o in objs}
        if objsdict:
            objs = [o for o in objs if o in six.iterkeys(objsdict)]
        for obj in objs:
            if obj in six.iterkeys(self.data):
                resp[obj] = True
                self.data[obj] -= 1
                if self.data[obj] <= 0:
                    del self.data[obj]
        self.update_monitor(objsdict)
        return resp

    def update_monitor(self, objsdict=None):
        if objsdict:
            regs = list(objsdict.keys())
            self.monitor = [k for k in six.iterkeys(self.data) if k in regs]
        else:
            self.monitor = list(self.data.keys())

    def get_monitor(self):
        return self.monitor


class SimInterface(object):
    all_interfaces = {}

    def __init__(self):
        pass

    def __call__(self, fun):
        #argspec = inspect.getargspec(fun)
        #argspec = inspect.formatargspec(*argspec)
        argspec = str(inspect.signature(fun))
        temp = argspec.split(',')
        interf = {
            'args': '(' + ', '.join(temp[1:]).strip(),
            'doc': fun.__doc__
        }
        SimInterface.all_interfaces[
            fun.__name__] = interf  #'('+', '.join(temp[1:]).strip()
        return fun


class SimCommand(object):
    sim_units = {
        'fs': BSM_FS,
        'ps': BSM_PS,
        'ns': BSM_NS,
        'us': BSM_US,
        'ms': BSM_MS,
        's': BSM_SEC
    }
    trigger_type = {
        'posneg': BSM_BOTHEDGE,
        'pos': BSM_POSEDGE,
        'neg': BSM_NEGEDGE,
        'none': BSM_NONEEDGE,
        BSM_BOTHEDGE: BSM_BOTHEDGE,
        BSM_POSEDGE: BSM_POSEDGE,
        BSM_NEGEDGE: BSM_NEGEDGE,
        BSM_NONEEDGE: BSM_NONEEDGE
    }
    tfile_format = {
        'bsm': BSM_TRACE_SIMPLE,
        'vcd': BSM_TRACE_VCD,
        BSM_TRACE_SIMPLE: BSM_TRACE_SIMPLE,
        BSM_TRACE_VCD: BSM_TRACE_VCD
    }

    def __init__(self, qcmd, qresp):
        self.simengine = None
        self.monitor = MonitorList()
        self.breakpoint = BpList()
        self.bp_values_prev = {}
        self.tfile = {}
        self.tfile_raw = {}
        self.tbuf = {}
        self.qcmd = qcmd
        self.qresp = qresp
        self.running = False
        self.sim_step = 20
        self.sim_unit_step = BSM_NS
        self.sim_total = -1  # -1 --> run forever
        self.sim_unit_total = BSM_NS
        self.sim_total_sec = -1

    def __del__(self):
        if self.is_valid():
            self.simengine.ctx_stop()

    def response(self, resp):
        self.qresp.put(resp)

    def check_bp(self, n):
        """check the breakpoints"""
        bps = self.breakpoint.get_bp()
        if not bps:
            return 0
        objs = list(bps.keys())
        values = self.read(objects=objs)
        for name, bp in six.iteritems(bps):
            t = bp.triggered(values[name], self.bp_values_prev[name])
            if t is not None:
                # set the important filed to be True, so the client will
                # never ignore the events
                resp = {
                    'cmd': 'breakpoint_triggered',
                    'important': True,
                    'value': [name, t[0], t[1], t[2]]
                }
                self.response(resp)
                self.running = False
                self.bp_values_prev.update(values)
                return 1
        self.bp_values_prev.update(values)
        return 0

    @SimInterface()
    def is_valid(self, **kwargs):
        if self.simengine and self.simengine.is_valid():
            return True
        return False

    @SimInterface()
    def load(self, filename="", **kwargs):
        if not os.path.exists(filename):
            return None
        self.simengine = SimSysC(filename)
        if self.is_valid():
            print(self.simengine.ctx.version)
            print(self.simengine.ctx.copyright)
            self.simengine.ctx_set_callback(self.check_bp)
            objs = {}
            for name, obj in six.iteritems(self.simengine.sim_objects):
                objs[name] = {
                    'name': obj.name,
                    'basename': obj.basename,
                    'kind': obj.kind,
                    'value': "",
                    'writable': obj.writable,
                    'readable': obj.readable,
                    'numeric': obj.numeric,
                    'parent': obj.parent,
                    'nkind': obj.nkind,
                    'register': obj.register
                }
            return objs
        return None

    @SimInterface()
    def step(self, running=False, **kwargs):
        """
        proceed the simulation with one step

        The step is set with set_parameter(). The GUI components will be updated
        after the running.

        The breakpoints are checked at each delta cycle.
        """
        if not self.is_valid():
            return False

        sim_step = self.sim_step
        sim_unit = self.sim_unit_step
        if self.sim_total_sec > 0:
            # not run infinitely
            t = self.sim_total_sec - self.simengine.ctx_time()
            if t > 0:
                # step should not exceed the remaining simulation time
                scale = [1e15, 1e12, 1e9, 1e6, 1e3, 1e0]
                sim_step = min(sim_step, int(t * scale[sim_unit] + 0.5))
            else:
                # finished, no need to run
                self.running = False
                return False
        self.running = (running and (sim_step == self.sim_step))
        self.simengine.ctx_start(sim_step, sim_unit)

        return True

    @SimInterface()
    def pause(self, **kwargs):
        """pause the simulation"""
        self.running = False
        return True

    def _parse_time(self, t):
        """
        parse the time in time+unit format

        For example,
            1) 1.5us will return (1.5, BSM_US)
            2) 100 will return (100, None), where unit is None (current one will
               be used)
        """
        pattern = r"([+-]?\d+(?:\.\d+)?(?:[eE][+-]?\d+)?)(?:\s)*(fs|ps|ns|us|ms|s|)"
        x = re.match(pattern, str(t))
        if x:
            if x.group(2):
                unit = self.sim_units.get(x.group(2), None)
                if unit is None:
                    raise ValueError("unknown time format: " + str(t))
                return float(x.group(1)), unit
            else:
                return float(x.group(1)), None
        return None, None

    @SimInterface()
    def set_parameter(self,
                      step=None,
                      total=None,
                      to=None,
                      more=None,
                      **kwargs):
        """set the parameters"""
        step, step_unit = self._parse_time(step)
        total, total_unit = self._parse_time(total)
        more, more_unit = self._parse_time(more)
        to, to_unit = self._parse_time(to)

        # global settings
        if total_unit is not None:
            self.sim_unit_total = total_unit
        if total is not None:
            self.sim_total = total
        if step_unit is not None:
            self.sim_unit_step = step_unit
        if step is not None:
            self.sim_step = step

        if to_unit is None:
            # unit not set, use the total unit
            to_unit = self.sim_unit_total
        if more_unit is None:
            more_unit = self.sim_unit_total

        scale = [1e15, 1e12, 1e9, 1e6, 1e3, 1e0]
        if self.sim_total > 0:
            self.sim_total_sec = self.sim_total / scale[self.sim_unit_total]
        else:
            self.sim_total_sec = self.sim_total

        if to is not None and to > 0:
            # calculate the next pause time in sec, no need to update the
            # absolute max simulation time (sim_total)
            to = to / scale[to_unit]
            # update the next pause time
            if self.sim_total_sec > 0:
                self.sim_total_sec = min(to, self.sim_total_sec)
            else:
                self.sim_total_sec = to
        elif more is not None and more > 0:
            more = more / scale[more_unit] + self.time_stamp(insecond=True)
            # update the next pause time
            if self.sim_total_sec > 0:
                self.sim_total_sec = min(more, self.sim_total_sec)
            else:
                self.sim_total_sec = more

        return self.get_parameter()

    @SimInterface()
    def get_parameter(self, **kwargs):
        return {
            'step_unit': self.sim_unit_step,
            'step': self.sim_step,
            'total': self.sim_total,
            'total_unit': self.sim_unit_total
        }

    def _object_list(self, objs):
        """help function to generate object list"""
        if isinstance(objs, six.string_types):
            return [objs]
        elif isinstance(objs, (list, tuple)):
            return objs
        else:
            raise ValueError()

    @SimInterface()
    def read(self, objects=None, **kwargs):
        """
        get the values of the registers

        If objects only contains one register, its value will be returned if
        succeed; otherwise a dictionary is returned, where the keys are the
        items in objects.

        Example: read a single register
        >>> read('top.sig_bool', True)

        Example: read multiple registers from the same simulation
        >>> read(['top.sig_bool', 'top.sig_cos']
        """
        if not self.is_valid():
            return {}

        if not objects:
            # empty list, read all the monitored objects
            objects = self.monitor.get_monitor()
        else:
            objects = self._object_list(objects)

        values = {}
        for obj in objects:
            values[obj] = self.simengine.ctx_read(obj)
        return values

    @SimInterface()
    def write(self, objects=None, **kwargs):
        """
        write the value to the registers

        objs should be a dictionary where the keys are the register name.
        Due to the two-step mechanism in SystemC, the value will be updated
        after the next delta cycle. That is, if a read() is called after
        write(), it will return the previous value.

        Example:
        >>> a = read('top.sig_int', True)
        >>> write({'top.sig_int': 100}, True)
        >>> b = read('top.sig_int', True) # a == b
        >>> step()
        >>> c = read('top.sig_int', True)
        """
        if not self.is_valid():
            return False
        if not objects or not isinstance(objects, dict):
            return False
        resp = {}
        for name, value in six.iteritems(objects):
            resp[name] = self.simengine.ctx_write(name, value)
        return resp

    @SimInterface()
    def time_stamp(self, insecond=False, **kwargs):
        """return the current simulation time stamp in string"""
        if not self.is_valid():
            return 0.0
        if insecond:
            #in second
            return self.simengine.ctx_time()
        return self.simengine.ctx_time_str()

    @SimInterface()
    def trace_file(self,
                   filename='',
                   name='',
                   fmt='bsm',
                   valid=None,
                   trigger='posneg',
                   **kwargs):
        """
        dump object values to a file

        filename:
            trace filename
        name:
            register name
        fmt:
            'bsm': only output the register value, one per line (Default)
            'vcd': output the SystemC VCD format data
        valid:
            the trigger signal. If it is none, the write-operation will be
            triggered by the register itself
        trigger:
            'posneg': trigger on both rising and falling edges
            'pos': trigger on rising edge
            'neg': trigger on falling edge
            'none': no triggering
        """
        if not self.is_valid():
            return False
        raw = [filename, name, fmt, valid, trigger]

        if not filename:
            filename = name
        if filename in self.tfile:
            return False

        obj = self.simengine.find_object(name)
        if obj is None:
            return False
        if valid:
            valid = self.simengine.find_object(valid)
            if valid is None:
                return False
            valid = valid()
        fmt = self.tfile_format.get(fmt, None)
        trigger = self.trigger_type.get(trigger, None)
        if fmt is None:
            raise ValueError("Not supported trace type: " + str(raw[0]))
        if trigger is None:
            raise ValueError("Not supported trigger type: " + str(raw[2]))
        trace = csim.SStructWrapper(self.simengine.csim.sim_trace_file())
        trace.name = filename
        trace.type = fmt
        if self.simengine.ctx_create_trace_file(trace()):
            self.simengine.ctx_trace_file(trace(), obj(), valid, trigger)
            self.tfile[filename] = {'trace': trace, 'raw': raw}
            return True

        return False

    @SimInterface()
    def close_trace_file(self, filename="", **kwargs):
        """stop dumping to a file"""
        if filename not in self.tfile:
            return False
        trace = self.tfile[filename]['trace']
        if self.simengine.ctx_close_trace_file(trace()):
            del self.tfile[filename]
            return True
        return False

    @SimInterface()
    def get_trace_files(self, **kwargs):
        """return all the trace files"""
        return {b: self.tfile[b]['raw'] for b in self.tfile}

    @SimInterface()
    def trace_buf(self,
                  name='',
                  size=256,
                  valid=None,
                  trigger="posneg",
                  **kwargs):
        """start dumping the register to a numpy array"""
        if not self.is_valid():
            return False
        # used to return the traced buffer list
        raw = [size, valid, trigger]
        obj = self.simengine.find_object(name)
        if obj is None:
            return False

        if valid:
            valid = self.simengine.find_object(valid)
            if valid is None:
                return False
            valid = valid()

        trigger = self.trigger_type.get(trigger, None)
        if trigger is None:
            raise ValueError("Not supported trigger type: " + str(raw[2]))

        if name in self.tbuf:
            # remove the existing trace
            trace = self.tbuf[name]['trace']
            self.simengine.ctx_close_trace_buf(trace())
            del self.tbuf[name]

        trace = csim.SStructWrapper(self.simengine.csim.sim_trace_buf())
        trace.name = name
        trace.size = size
        data = np.zeros((size))
        trace.buffer = data.ctypes.data_as(ctypes.POINTER(ctypes.c_double))
        if self.simengine.ctx_create_trace_buf(trace()):
            self.simengine.ctx_trace_buf(trace(), obj(), valid, trigger)
            self.tbuf[name] = {'trace': trace, 'data': data, 'raw': raw}
        return True

    @SimInterface()
    def close_trace_buf(self, name='', **kwargs):
        """stop dumping to a numpy array"""
        if name not in self.tbuf:
            return False

        trace = self.tbuf[name]['trace']
        if self.simengine.ctx_close_trace_buf(trace()):
            del self.tbuf[name]
            return True
        return False

    @SimInterface()
    def read_buf(self, objects=None, **kwargs):
        """
        read the traced buffer to an numpy array

        If the buffer is previous traced by calling trace_buf, the array with
        previous defined size will return; otherwise the trace_buf will be
        called with default arguments first.
        """
        if not self.is_valid():
            return {}
        if not objects:
            # no object defined, return all the traced buffers
            objects = list(self.tbuf.keys())
        else:
            objects = self._object_list(objects)

        resp = {}
        for name in objects:
            obj = self.simengine.find_object(name)
            if obj is None:
                continue

            if name not in self.tbuf:
                # if not traced yet, trace it now with default settings
                self.trace_buf(name=name)

            self.simengine.ctx_read_trace_buf(self.tbuf[name]['trace']())
            resp[name] = self.tbuf[name]['data']
        return resp

    @SimInterface()
    def get_trace_bufs(self, **kwargs):
        return {b: self.tbuf[b]['raw'] for b in self.tbuf}

    @SimInterface()
    def monitor_signal(self, objects=None, **kwargs):
        """
        monitor the register value

        At end of each step, the simulation process will report the value
        """
        if not self.is_valid() or not objects:
            return {}

        objects = self._object_list(objects)
        return self.monitor.add(objects, self.simengine.sim_objects)

    @SimInterface()
    def unmonitor_signal(self, objects=None, **kwargs):
        """stop monitoring the register"""
        if not self.is_valid() or not objects:
            return {}

        objects = self._object_list(objects)
        return self.monitor.delete(objects, self.simengine.get_objects())

    @SimInterface()
    def get_monitored_signals(self, **kwargs):
        """get the list of the monitored signals"""
        return self.monitor.get_monitor()

    @SimInterface()
    def get_status(self, **kwargs):
        return {'valid': self.is_valid(), 'running': self.running}

    @SimInterface()
    def add_breakpoint(self, name="", condition=None, hitcount=None, **kwargs):
        """
        add the breakpoint

        bp = (name, condition, hitcount) or name
        """
        if not self.is_valid():
            return {}
        if not self.simengine.find_object(name):
            raise ValueError("Invalid object %s" % name)
        objs = self.breakpoint.add([[name, condition, hitcount]],
                                   self.simengine.get_objects())
        bps = self.breakpoint.get_bp()
        # update the current breakpoint values before they are
        # checked next time
        self.bp_values_prev = self.read(objects=list(bps.keys()))
        return objs

    @SimInterface()
    def del_breakpoint(self, name="", condition=None, hitcount=None, **kwargs):
        """delete the breakpoint"""
        if not self.is_valid():
            return {}

        if not self.simengine.find_object(name):
            raise ValueError("Invalid object %s" % name)
        return self.breakpoint.delete([[name, condition, hitcount]],
                                      self.simengine.get_objects())

    @SimInterface()
    def get_breakpoints(self, **kwargs):
        """get all the breakpoints"""
        return self.breakpoint.data_raw

    @SimInterface()
    def get_interfaces(self, **kwargs):
        return SimInterface.all_interfaces

    def process(self):
        while True:
            running = self.running
            if self.running:
                if not self.is_valid():
                    self.running = False
                else:
                    remain = self.sim_total_sec - self.simengine.ctx_time()
                    if not ((self.sim_total_sec <= 0) or (remain > 0)):
                        # pause simulation when reach the simulation time
                        self.running = False
            try:
                if self.running:
                    self.step(running=self.running)
                if self.running:
                    # In running mode, return immediately so that the simulation
                    # can be triggered as soon as possible
                    cmd = self.qcmd.get_nowait()
                else:
                    cmd = self.qcmd.get()
                if not cmd:
                    continue

                command = cmd.get('cmd', '')
                args = cmd.get('arguments', {})

                # the response sends back the original command and one more
                # field, i.e., 'value'
                resp = cmd
                resp['value'] = False
                if not command and not args:
                    raise ValueError("unknown command: ", cmd)

                if command == 'exit':
                    return False

                if not args.get('silent', True):
                    print(cmd)

                if hasattr(self, command):
                    fun = getattr(self, command)
                    resp['value'] = fun(**args)
                else:
                    raise ValueError('unknown command: ', cmd)

                self.response(resp)
                if self.running != running:
                    # automatically report status change
                    self.response({
                        'cmd': 'get_status',
                        'value': self.get_status()
                    })
            except Queue.Empty:
                pass
            except:
                if resp:
                    self.response(resp)
                traceback.print_exc(file=sys.stdout)


class SimLogger(object):
    def __init__(self, qresp):
        self.qresp = qresp

    def write(self, buf):
        self.qresp.put({'cmd': 'write_out', 'important': True, 'value': buf})

    def flush(self):
        pass


def sim_process(qresp, qcmd):
    log = SimLogger(qresp)
    stdout = sys.stdout
    stderr = sys.stderr
    sys.stdout = log
    sys.stderr = log
    proc = SimCommand(qcmd, qresp)
    # infinite loop
    proc.process()
    sys.stdout = stdout
    sys.stderr = stderr
